import unittest
from typing import Type

from nichtparasoup.imagecrawler import BaseImageCrawler
from nichtparasoup.imagecrawler.echo import Echo
from nichtparasoup.testing.imagecrawler import ImageCrawlerLoaderTest


class EchoConfigImageUriTest(unittest.TestCase):

    def test__check_config_right_value(self) -> None:
        # arrange
        config_in = dict(image_uri="test")
        # act
        config_out = Echo.check_config(config_in)
        # assert
        self.assertDictEqual(config_in, config_out)

    def test__check_config_missing_value(self) -> None:
        # assert
        with self.assertRaises(KeyError):
            Echo.check_config(dict())

    def test__check_config_wrong_type(self) -> None:
        wrong_types = [None, True, 23, 4.2, [], (), {}, self]  # type: ignore
        for wrong_type in wrong_types:
            # assert
            with self.assertRaises(TypeError, msg=repr(wrong_type)):
                Echo.check_config(dict(image_uri=wrong_type))

    def test__check_config_wrong_value(self) -> None:
        wrong_values = [""]
        for wrong_value in wrong_values:
            # assert
            with self.assertRaises(ValueError, msg=repr(wrong_value)):
                Echo.check_config(dict(image_uri=wrong_value))


class EchoCrawlTest(unittest.TestCase):

    def test_crawl(self) -> None:
        # arrange
        crawler = Echo(image_uri="test")
        # act
        images_crawled = crawler.crawl()
        images_crawled_len = len(images_crawled)
        image_crawled = images_crawled.pop() if images_crawled_len else None
        # assert
        self.assertEqual(images_crawled_len, 1, "no images crawled")
        if image_crawled:
            self.assertTrue(image_crawled.is_generic, "this is not a generic")
            self.assertTrue(image_crawled.more.get("this_is_a_dummy"), "this is not a dummy")


class EchoDescriptionTest(unittest.TestCase):

    def test_description_config(self) -> None:
        # act
        description = Echo.info()
        # assert
        assert isinstance(description.config, dict)
        self.assertIn('image_uri', description.config)


class EchoLoaderTest(ImageCrawlerLoaderTest):

    @property
    def ic_name(self) -> str:
        return 'Echo'

    @property
    def ic_class(self) -> Type[BaseImageCrawler]:
        return Echo

    def test_loader(self) -> None:
        self.check()



__all__ = ["Echo"]

from typing import Any, Dict

from nichtparasoup.imagecrawler import BaseImageCrawler, Image, ImageCollection, ImageCrawlerConfig, ImageCrawlerInfo


class Echo(BaseImageCrawler):

    @classmethod
    def info(cls) -> ImageCrawlerInfo:
        from textwrap import dedent
        return ImageCrawlerInfo(
            description='"Finds" the same image ... again ... and again.',
            long_description=dedent('''
                Not an actual crawler.
                More like an Parrot that is trained to repeat what you tell it to say.
                ''').strip(),
            config=dict(
                image_uri='the URI of the image to "find"',
            ),
            # does not have an icon
        )

    @classmethod
    def check_config(cls, config: Dict[Any, Any]) -> ImageCrawlerConfig:
        image_uri = config["image_uri"]
        if type(image_uri) is not str:
            raise TypeError("image_uri {!r} is not str".format(image_uri))
        if 0 == len(image_uri):
            raise ValueError("image_uri {!r} is empty".format(image_uri))
        return ImageCrawlerConfig(
            image_uri=image_uri,
        )

    def _reset(self) -> None:  # pragma: no cover
        pass

    def _crawl(self) -> ImageCollection:
        images = ImageCollection()
        config = self.get_config()
        images.add(Image(
            uri=config["image_uri"],
            source=config["image_uri"],
            is_generic=True,
            this_is_a_dummy=True,
        ))
        return images



__all__ = ["Image", "ImageCollection"]

from typing import Any, Set
from uuid import uuid4

ImageUri = str

SourceUri = str


class Image(object):
    """
    Describe an image

    `uri`
        The absolute URI of the image. This basically identifies the Image and makes it unique.

        This absolute URI must include: ``scheme``, ``host``.
            ``schema`` must be either 'http' or 'https' - use 'https' if possible!
        Optional are: ``port``, ``path``, ``query``, ``fragment``.

    `source`
        The URI where did the image is originally found?
        This URI can point to a ImageBoardThread, or a comment section in a Forum, or a news article...

        In the idea of fair use, it is encouraged to point to the source as good as possible.

        This absolute URI must include: ``scheme``, ``host``.
            ``schema`` must be either 'http' or 'https' - the last one is preferred.
        Optional are: ``port``, ``path``, ``query``, ``fragment``.

        Good examples are:
            * https://www.reddit.com/r/Awww/comments/e1er0c/say_hi_to_loki_hes_just_contemplating/
            * https://giphy.com/gifs/10kABVanhwykJW

    `is_generic`
        If a generic image crawler is used, its common that each image URI looks exactly the same.
        To make this known, use this flag.

    `more`
        A dictionary of additional information an image crawler might want to deliver.

        This dictionary's data types are intended to the basic ones: string, int, float, list, set, dict, bool, None

        Good examples are:
            * image-dimensions
            * author, copyright information
            * valid-until

    """

    def __init__(self, *,
                 uri: ImageUri, source: SourceUri,
                 is_generic: bool = False,
                 **more: Any) -> None:  # pragma: no cover
        self.uri = uri
        self.source = source
        self.more = more
        self.is_generic = is_generic
        self.__hash = hash(uuid4()) if self.is_generic else hash(self.uri)

    def __hash__(self) -> int:
        return self.__hash

    def __eq__(self, other: Any) -> bool:
        if type(other) is type(self):
            return hash(self) == hash(other)
        return False

    def __repr__(self) -> str:  # pragma: no cover
        return '<{0.__module__}.{0.__name__} object at {1:#x} {2.uri!r}>'.format(type(self), id(self), self)


class ImageCollection(Set[Image]):
    pass



__all__ = ["WebServer"]

from json import dumps as json_encode
from os.path import dirname, join as path_join
from typing import Any, Dict, Mapping, Optional, Sequence, Set, Tuple, Type, Union

from mako.template import Template  # type: ignore
from werkzeug.datastructures import Headers
from werkzeug.exceptions import HTTPException, NotFound
from werkzeug.routing import Map, Rule
from werkzeug.utils import redirect
from werkzeug.wrappers import Request, Response

from nichtparasoup.core.imagecrawler import BaseImageCrawler
from nichtparasoup.core.server import Server, ServerStatus, type_module_name_str


class JsonRespone(Response):
    def __init__(
        self,
        response: Optional[Any] = None,
        status: Optional[Union[str, int]] = None,
        headers: Optional[Union[Headers, Mapping[str, str], Sequence[Tuple[str, str]]]] = None,
        mimetype: Optional[str] = 'application/json',
        content_type: Optional[str] = 'application/json',
        direct_passthrough: bool = False,
    ) -> None:
        super().__init__(
            response=json_encode(response),
            status=status,
            headers=headers,
            mimetype=mimetype,
            content_type=content_type,
            direct_passthrough=direct_passthrough
        )


class WebServer(object):
    _TEMPLATE_FILES = path_join(dirname(__file__), 'htdocs', 'template')
    _STATIC_FILES = path_join(dirname(__file__), 'htdocs', 'static')
    _STATIC_INDEX = 'index.html'  # relative to cls._STATIC_FILES

    def __init__(self, imageserver: Server, hostname: str, port: int) -> None:  # pragma: no cover
        self.imageserver = imageserver
        self.hostname = hostname
        self.port = port
        self.url_map = Map([
            Rule('/', endpoint='root'),
            Rule('/get', endpoint='get'),
            Rule('/status', endpoint='status'),
            Rule('/status/<what>', endpoint='status_what'),
            Rule('/reset', endpoint='reset'),
            Rule('/css/sourceIcons.css', endpoint='sourceicons')
        ])

    def __call__(self, environ: Dict[str, Any], start_response: Any) -> Any:
        return self.wsgi_app(environ, start_response)

    def dispatch_request(self, request: Request) -> Union[Response, HTTPException]:
        adapter = self.url_map.bind_to_environ(request.environ)
        try:
            endpoint, values = adapter.match()
            response = getattr(self, 'on_{}'.format(endpoint))(request, **values)  # type: Response
            return response
        except HTTPException as e:
            return e

    def wsgi_app(self, environ: Dict[str, Any], start_response: Any) -> Any:
        request = Request(environ)
        response = self.dispatch_request(request)
        if isinstance(response, Response):
            response.cache_control.no_cache = True
            response.cache_control.no_store = True
        return response(environ, start_response)

    def on_root(self, _: Request) -> Response:
        # relative-path is valid via https://tools.ietf.org/html/rfc3986#section-4.2
        forward = redirect(self._STATIC_INDEX, code=302, Response=Response)
        # to prevent extensive (reverse proxy) header parsing, it is kept as a relative-path
        forward.autocorrect_location_header = False
        return forward

    def on_get(self, _: Request) -> Response:
        image = self.imageserver.get_image()
        return JsonRespone(image)

    _STATUS_WHATS = dict(
        server=ServerStatus.server,
        blacklist=ServerStatus.blacklist,
        crawlers=ServerStatus.crawlers,
    )

    def on_status(self, _: Request) -> Response:
        status = {what: getter(self.imageserver) for what, getter in self._STATUS_WHATS.items()}
        return JsonRespone(status)

    def on_status_what(self, _: Request, what: str) -> Response:
        status_what = self._STATUS_WHATS.get(what)
        if not status_what:
            raise NotFound()
        status = status_what(self.imageserver)
        return JsonRespone(status)

    def on_reset(self, _: Request) -> Response:
        reset = self.imageserver.request_reset()
        return JsonRespone(reset)

    def on_sourceicons(self, _: Request) -> Response:
        imagecrawlers = {
            type(crawler.imagecrawler)
            for crawler
            in self.imageserver.core.crawlers}  # type: Set[Type[BaseImageCrawler]]
        names_icons_list = [
            (name, icon)
            for name, icon
            in [
                (type_module_name_str(imagecrawler), imagecrawler.info().icon_url)
                for imagecrawler
                in imagecrawlers]
            if icon]
        # cannot use dict for `names_icons_list` in template. will break the template occasionally :-/
        template = Template(filename=path_join(self._TEMPLATE_FILES, 'css', 'sourceIcons.css'))
        css = template.render(names_icons_list=names_icons_list)
        return Response(css, mimetype='text/css')

    def run(self) -> None:
        from werkzeug.serving import run_simple
        from nichtparasoup._internals import _log
        self.imageserver.start()
        try:
            _log('info', ' * starting {0} bound to {1.hostname} on port {1.port}'.format(type(self).__name__, self))
            run_simple(
                self.hostname, self.port,
                application=self,
                static_files={'/': self._STATIC_FILES},
                processes=1, threaded=True,
                use_reloader=False,
                use_debugger=False)
            _log('info', ' * stopped {0} bound to {1.hostname} on port {1.port}'.format(type(self).__name__, self))
        except Exception as e:
            _log('exception', ' * Error occurred. stopping everything')
            raise e
        finally:
            self.imageserver.stop()



# pragma: no cover

__all__ = ["parser"]

from argparse import ArgumentParser

parser = ArgumentParser(
    add_help=True,
    allow_abbrev=False,
)

commands = parser.add_subparsers(
    title='Commands',
    metavar='<command>',
    dest='command',
)
commands.required = True

parser.add_argument(
    '--debug',
    help='enable debug output',
    action='store_true', dest="debug",
)

command_run = commands.add_parser(
    'run',
    help='run a server',
    description='start a webserver to display random images',
    add_help=True,
    allow_abbrev=False,
)
command_run.add_argument(
    '-c', '--use-config',
    help='use a YAML config file instead of the defaults',
    metavar='<file>',
    action='store', dest="config_file", type=str,
)

command_config = commands.add_parser(
    'config',
    description='Get config related things done',
    help='config related functions',
    add_help=True,
    allow_abbrev=False,
)
command_config_actions = command_config.add_mutually_exclusive_group(required=True)
command_config_actions.add_argument(
    '--check',
    help='validate and probe a YAML config file',
    action='store_const', dest='action', const='check',
)
command_config_actions.add_argument(
    '--dump',
    help='dump YAML config into a file',
    action='store_const', dest='action', const='dump',
)
command_config.add_argument(
    metavar='<file>',
    action='store', dest='config_file', type=str,
)

command_info = commands.add_parser(
    'info',
    description='Get info for several topics',
    help='get info for several topics',
    add_help=True,
    allow_abbrev=False,
)
command_info_actions = command_info.add_mutually_exclusive_group(required=True)
command_info_actions.add_argument(
    '--imagecrawler-list',
    help='list available image crawler types',
    action='store_true', dest='imagecrawler_list',
)
command_info_actions.add_argument(
    '--imagecrawler-desc',
    help='describe an image crawler type and its config',
    metavar='<crawler>',
    action='store', dest='imagecrawler_desc', type=str,
)
command_info_actions.add_argument(
    '--version',
    help="show program's version number",
    action='store_true', dest='version',
)



import unittest
from typing import Any, Dict
from unittest.mock import MagicMock

from nichtparasoup.core import NPCore
from nichtparasoup.core.image import Image, ImageCollection
from nichtparasoup.core.server import Server, type_module_name_str

from .mockable_imagecrawler import MockableImageCrawler


class ServerGetImageTest(unittest.TestCase):

    def setUp(self) -> None:
        self.server = Server(NPCore())

    def tearDown(self) -> None:
        del self.server

    def test_get_image_no_crawler(self) -> None:
        # act
        image = self.server.get_image()
        # assert
        self.assertIsNone(image)

    def test_get_image_no_images(self) -> None:
        # arrange
        self.server.core.add_imagecrawler(MockableImageCrawler(), 1)
        # act
        image = self.server.get_image()
        # assert
        self.assertIsNone(image)

    def test_get_image_some_images(self) -> None:
        # arrange
        image_crawled = Image(uri='testURI', is_generic=True, source='testSource', bla=1, foo="bar")
        imagecrawler = MockableImageCrawler()
        imagecrawler.crawl = MagicMock(return_value=ImageCollection([image_crawled]))  # type: ignore
        self.server.core.add_imagecrawler(imagecrawler, 1)
        crawler = self.server.core.crawlers[0]
        # act
        self.server.core.crawlers[0].crawl()
        image_got = self.server.get_image()
        # assert
        self.assertIsInstance(image_got, dict)
        if image_got:  # needed to trick mypy
            self.assertEqual(image_got.get("uri"), image_crawled.uri)
            self.assertEqual(image_got.get("is_generic"), image_crawled.is_generic)
            self.assertEqual(image_got.get("source"), image_crawled.source)
            self.assertEqual(image_got.get("more"), image_crawled.more)
            self.assertIsInstance(image_got.get("crawler"), dict)
            image_got_crawler = image_got["crawler"]  # type: Dict[str, Any]
            self.assertEqual(image_got_crawler.get("id"), id(crawler))
            self.assertEqual(image_got_crawler.get("type"), type_module_name_str(type(imagecrawler)))


class ServerResetTest(unittest.TestCase):
    def setUp(self) -> None:
        self.server = Server(NPCore(), reset_timeout=7357)
        self.reset_called = False
        self.server._reset = self.mock__reset  # type: ignore

    def mock__reset(self) -> None:
        self.reset_called = True

    def tearDown(self) -> None:
        del self.reset_called
        del self.server

    def test_not_running(self) -> None:
        # act
        res = self.server.request_reset()
        # assert
        self.assertTrue(self.reset_called)
        self.assertDictEqual(res, dict(
            requested=True,
            timeout=0,
        ))

    def test_running_under_timeout(self) -> None:
        from time import time
        # arrange
        self.server.is_alive = lambda: True  # type: ignore
        self.server._stats.time_started = int(time())
        # act
        res = self.server.request_reset()
        # assert
        self.assertFalse(self.reset_called)
        self.assertDictEqual(res, dict(
            requested=False,
            timeout=self.server.reset_timeout,
        ))

    def test_running_over_timeout(self) -> None:
        from time import time
        # arrange
        self.server.is_alive = lambda: True  # type: ignore
        self.server._stats.time_started = int(time()) - 2 * self.server.reset_timeout
        # act
        res = self.server.request_reset()
        # assert
        self.assertTrue(self.reset_called)
        self.assertDictEqual(res, dict(
            requested=True,
            timeout=self.server.reset_timeout,
        ))


class ServerStartStopTest(unittest.TestCase):

    def setUp(self) -> None:
        self.server = Server(NPCore())

    def tearDown(self) -> None:
        try:
            self.server.stop()
        except Exception:
            pass
        finally:
            del self.server

    def test_start_unlocked(self) -> None:
        # act
        self.server.start()
        # assert
        self.assertFalse(self.server._locks.run.locked())

    def test_start_while_running_unlocked(self) -> None:
        # arrange
        self.server.start()
        # act
        with self.assertRaises(RuntimeError):
            self.server.start()
        # assert
        self.assertFalse(self.server._locks.run.locked())

    def test_stop_unlocked(self) -> None:
        # arrange
        self.server.start()
        # act
        self.server.stop()
        # assert
        self.assertFalse(self.server._locks.run.locked())

    def test_stop_while_not_running_unlocked(self) -> None:
        # act
        with self.assertRaises(RuntimeError):
            self.server.stop()
        # assert
        self.assertFalse(self.server._locks.run.locked())

    def test_start_stop_repeat(self) -> None:
        self.assertFalse(self.server.is_alive())
        self.server.start()
        self.assertTrue(self.server.is_alive())
        self.server.stop()
        self.assertFalse(self.server.is_alive())
        self.server.start()
        self.assertTrue(self.server.is_alive())
        self.server.stop()
        self.assertFalse(self.server.is_alive())



import unittest
from os import path
from pathlib import Path

from nichtparasoup.testing.imagecrawler import FileFetcher


class FileFetcherTest(unittest.TestCase):

    _TESTDATA_DIR = path.join(path.dirname(__file__), 'testdata_filefetcher')

    def test__uri_sort_query__dings(self) -> None:
        # assert
        self.assertEqual(
            'https://asdf',
            FileFetcher._uri_sort_query('https://asdf')
        )

    def test__uri_sort_query__path(self) -> None:
        # assert
        self.assertEqual(
            'https://as/df',
            FileFetcher._uri_sort_query('https://as/df')
        )

    def test__uri_sort_query__fragment(self) -> None:
        # assert
        self.assertEqual(
            'https://asdf#foo',
            FileFetcher._uri_sort_query('https://asdf#foo')
        )

    def test__uri_sort_query__no_query(self) -> None:
        # assert
        self.assertEqual(
            'https://asdf',
            FileFetcher._uri_sort_query('https://asdf?')
        )

    def test__valid_url(self) -> None:
        self.assertTrue(FileFetcher._valid_uri('file:///foo_bar/baz'))
        self.assertFalse(FileFetcher._valid_uri('http://foo.bar/baz'))
        self.assertFalse(FileFetcher._valid_uri('https://foo.bar/baz'))

    def test__uri_sort_query__empty_query(self) -> None:
        # assert
        self.assertEqual(
            'https://asdf?foo=',
            FileFetcher._uri_sort_query('https://asdf?foo=')
        )

    def test__uri_sort_query__query(self) -> None:
        # assert
        self.assertEqual(
            'https://asdf?foo=1',
            FileFetcher._uri_sort_query('https://asdf?foo=1')
        )

    def test__uri_sort_query__query_sorted(self) -> None:
        # assert
        self.assertEqual(
            'https://asdf?bar=1&foo=2',
            FileFetcher._uri_sort_query('https://asdf?bar=1&foo=2')
        )

    def test__uri_sort_query__unsorted(self) -> None:
        # assert
        self.assertEqual(
            'https://asdf?bar=2&foo=1',
            FileFetcher._uri_sort_query('https://asdf?foo=1&bar=2')
        )

    def test__get_file_uri__without_basedir(self) -> None:
        # arrange
        test_file = 'test.txt'
        # act & assert
        with self.assertRaisesRegex(FileNotFoundError, 'Path not absolute'):
            FileFetcher._build_uri(test_file)

    def test__get_file_uri__unknown_file(self) -> None:
        # arrange
        test_file = path.join(self._TESTDATA_DIR, 'unknown.file')
        # act & assert
        with self.assertRaisesRegex(FileNotFoundError, 'Not a file'):
            FileFetcher._build_uri(test_file)

    def test__get_file_uri__absolute_basedir(self) -> None:
        # arrange
        test_file = path.join(self._TESTDATA_DIR, 'test.txt')
        # act
        file = FileFetcher._build_uri(test_file)
        # assert
        self.assertTrue(file.startswith('file://'), file)
        self.assertEqual(
            Path(test_file).as_uri(),
            file
        )

    def test__get_file_uri__with_basedir(self) -> None:
        # arrange
        test_file = 'test.txt'
        test_base = self._TESTDATA_DIR
        # act
        file = FileFetcher._build_uri(test_file, test_base)
        # assert
        self.assertTrue(file.startswith('file://'), file)
        self.assertEqual(
            Path(path.join(test_base, test_file)).as_uri(),
            file,
        )

    def test__get_file_uri__unknown_basedir(self) -> None:
        # arrange
        filefetcher = FileFetcher(dict())
        # act & assert
        with self.assertRaises(FileNotFoundError):
            filefetcher._get_file_uri('unknown_file')

    def test_get_stream(self) -> None:
        from urllib.response import addinfourl
        # arrange
        filefetcher = FileFetcher(
            dict(
                test_file='test.txt',
            ),
            base_dir=self._TESTDATA_DIR
        )
        # act
        stream, _ = filefetcher.get_stream('test_file')
        # assert
        self.assertIsInstance(stream, addinfourl)



import unittest
from os import path

from pkg_resources import Distribution, EntryPoint

from nichtparasoup.imagecrawler import KnownImageCrawlers

from .testdata_imagecrawler_plugin import (
    BaseLoadableImageCrawlerA, LoadableImageCrawlerA, LoadableImageCrawlerB, StandaloneLoadableImageCrawlerA,
)

_TEST_PLUGIN_DIST = Distribution(location=path.abspath('testdata_imagecrawler_plugin'),
                                 project_name='testdata-imagecrawler-plugin')
_TEST_PLUGIN_CLASS = LoadableImageCrawlerA
_TEST_PLUGIN_ENTRY = EntryPoint('Test', _TEST_PLUGIN_CLASS.__module__, attrs=(_TEST_PLUGIN_CLASS.__name__,),
                                dist=_TEST_PLUGIN_DIST)


class KnownImageCrawlersLoadTest(unittest.TestCase):

    def test_success(self) -> None:
        # arrange
        entry = _TEST_PLUGIN_ENTRY
        # act
        loaded = KnownImageCrawlers._load(entry)
        # assert
        self.assertIs(_TEST_PLUGIN_CLASS, loaded)

    def test_raise(self) -> None:
        # arrange
        entry = EntryPoint('Test', '__.does.not.exist', attrs=('UnknownClass',),
                           dist=_TEST_PLUGIN_DIST)
        # act & assert
        with self.assertRaises(ImportError):
            KnownImageCrawlers._load(entry)


class KnownImageCrawlersInheritanceTest(unittest.TestCase):

    def test_success(self) -> None:
        # act
        KnownImageCrawlers._test_inheritance(LoadableImageCrawlerA)
        # assert
        self.assertTrue(True, 'no errors thrown')

    def test_raise(self) -> None:
        # act & assert
        with self.assertRaises(TypeError):
            KnownImageCrawlers._test_inheritance(StandaloneLoadableImageCrawlerA)


class KnownImageCrawlersAbstractTest(unittest.TestCase):

    def test_success(self) -> None:
        # act
        KnownImageCrawlers._test_abstract(StandaloneLoadableImageCrawlerA)
        # assert
        self.assertTrue(True, 'no errors thrown')

    def test_success_inherited(self) -> None:
        # act
        KnownImageCrawlers._test_abstract(LoadableImageCrawlerA)
        # assert
        self.assertTrue(True, 'no errors thrown')

    def test_raise(self) -> None:
        # act & assert
        with self.assertRaises(TypeError):
            KnownImageCrawlers._test_abstract(BaseLoadableImageCrawlerA)


class KnownImageCrawlersDuplicateNameTest(unittest.TestCase):

    def test_success_empty(self) -> None:
        # arrange
        kic = KnownImageCrawlers([])
        kic._list = []
        # act
        kic._test_duplicate_name('test')
        # assert
        self.assertTrue(True, 'no errors thrown')

    def test_success_filled(self) -> None:
        # arrange
        kic = KnownImageCrawlers([])
        kic._list = [('testA', LoadableImageCrawlerA)]
        # act
        kic._test_duplicate_name('testB')
        # assert
        self.assertTrue(True, 'no errors thrown')

    def test_raise(self) -> None:
        # arrange
        known_ic_name = 'test'
        kic = KnownImageCrawlers([])
        kic._list = [(known_ic_name, LoadableImageCrawlerA)]
        # act & assert
        with self.assertRaises(KeyError):
            kic._test_duplicate_name(known_ic_name)


class KnownImageCrawlersDuplicateClassTest(unittest.TestCase):

    def test_success_empty(self) -> None:
        # arrange
        kic = KnownImageCrawlers([])
        kic._list = []
        # act
        kic._test_duplicate_class(LoadableImageCrawlerA)
        # assert
        self.assertTrue(True, 'no errors thrown')

    def test_success_filled(self) -> None:
        # arrange
        kic = KnownImageCrawlers([])
        kic._list = [('test', LoadableImageCrawlerA)]
        # act
        kic._test_duplicate_class(LoadableImageCrawlerB)
        # assert
        self.assertTrue(True, 'no errors thrown')

    def test_raise(self) -> None:
        # arrange
        known_ic_class = LoadableImageCrawlerA
        kic = KnownImageCrawlers([])
        kic._list = [('test', known_ic_class)]
        # act & assert
        with self.assertRaises(TypeError):
            kic._test_duplicate_class(known_ic_class)


class KnownImageCrawlersTest(unittest.TestCase):

    def test_names(self) -> None:
        # arrange
        kic = KnownImageCrawlers([])
        kic._list = [('test', LoadableImageCrawlerA)]
        # act
        ic_names = kic.names()
        # arrange
        self.assertListEqual(['test'], [n for n in ic_names])

    def test_classes(self) -> None:
        # arrange
        kic = KnownImageCrawlers([])
        kic._list = [('test', LoadableImageCrawlerA)]
        # act
        ic_classes = kic.classes()
        # arrange
        self.assertListEqual([LoadableImageCrawlerA], [c for c in ic_classes])

    def test_items(self) -> None:
        # arrange
        kic = KnownImageCrawlers([])
        kic._list = [('test', LoadableImageCrawlerA)]
        # act
        ic_items = kic.items()
        # arrange
        self.assertListEqual(kic._list, [i for i in ic_items])

    def test_get_name(self) -> None:
        # arrange
        kic = KnownImageCrawlers([])
        kic._list = [('testKnown', LoadableImageCrawlerA)]
        # act
        ic_name_known = kic.get_name(LoadableImageCrawlerA)
        ic_name_unknown = kic.get_name(LoadableImageCrawlerB)
        # arrange
        self.assertEqual('testKnown', ic_name_known)
        self.assertIsNone(ic_name_unknown)

    def test_get_class(self) -> None:
        # arrange
        kic = KnownImageCrawlers([])
        kic._list = [('testKnown', LoadableImageCrawlerA)]
        # act
        ic_class_known = kic.get_class('testKnown')
        ic_class_unknown = kic.get_class('testUnknown')
        # arrange
        self.assertIs(LoadableImageCrawlerA, ic_class_known)
        self.assertIsNone(ic_class_unknown)


class KnownImageCrawlersAddTest(unittest.TestCase):

    def test_success(self) -> None:
        kic = KnownImageCrawlers([])
        kic._list = []
        # act
        kic._add(_TEST_PLUGIN_ENTRY)
        # arrange
        self.assertListEqual([(_TEST_PLUGIN_ENTRY.name, _TEST_PLUGIN_CLASS)], kic._list)


class KnownImageCrawlersBuiltinsTest(unittest.TestCase):

    def test_builtins(self) -> None:
        """ test each builtin for the usual plugin needs """
        builtins = KnownImageCrawlers._builtins()
        for _, bi_class in builtins.items():
            KnownImageCrawlers._test(bi_class)
        self.assertTrue(True, 'no exceptions thrown')



__all__ = ["NullCrawler", "Random3Crawler"]

from typing import Optional

from nichtparasoup.core import Crawler, _CrawlerWeight, _IsImageAddable, _OnImageAdded
from nichtparasoup.core.image import Image
from nichtparasoup.core.imagecrawler import BaseImageCrawler


class _LoggingCrawler(Crawler):
    def __init__(self,
                 imagecrawler: BaseImageCrawler, weight: _CrawlerWeight,
                 is_image_addable: Optional[_IsImageAddable] = None,
                 on_image_added: Optional[_OnImageAdded] = None) -> None:
        super().__init__(imagecrawler, weight, is_image_addable, on_image_added)
        self.crawl_call_count = 0
        self.reset_call_count = 0

    def crawl(self) -> int:
        self.crawl_call_count += 1
        return 0

    def reset(self) -> None:
        self.reset_call_count += 1


class NullCrawler(_LoggingCrawler):
    pass


class Random3Crawler(_LoggingCrawler):

    def crawl(self) -> int:
        super().crawl()
        for _ in range(3):
            self.images.add(Image(uri='test', source='test', is_generic=True))
        return 3



import unittest
from unittest.mock import MagicMock
from weakref import ref as weak_ref

from nichtparasoup.core import Crawler
from nichtparasoup.core.image import Image, ImageCollection

from .mockable_imagecrawler import MockableImageCrawler


class _C(object):

    def m_b(self, _: Image) -> bool:
        return True

    def m(self, _: Image) -> None:
        pass


def _f_b(_: Image) -> bool:
    return True


def _f(_: Image) -> None:
    pass


class CrawlerIsImageAddableTestCase(unittest.TestCase):
    """
    test if the weakref is working as expected
    """

    def setUp(self) -> None:
        self.crawler = Crawler(MockableImageCrawler(), 1)

    def tearDown(self) -> None:
        del self.crawler

    def test_default(self) -> None:
        # assert
        self.assertIsNone(self.crawler.get_is_image_addable())

    def test_none(self) -> None:
        # arrange
        obj = _C()
        self.crawler.set_is_image_addable(obj.m_b)
        # act
        self.crawler.set_is_image_addable(None)
        # assert
        self.assertIsNone(self.crawler.get_is_image_addable())

    def test_object_bound_method_stay(self) -> None:
        # arrange
        obj = _C()
        # act
        self.crawler.set_is_image_addable(obj.m_b)
        # assert
        self.assertEqual(self.crawler.get_is_image_addable(), obj.m_b)

    def test_object_bound_method_lambda(self) -> None:
        # arrange
        obj = _C()
        self.crawler.set_is_image_addable(obj.m_b)
        # act
        self.crawler.set_is_image_addable(_C().m_b)
        # assert
        self.assertIsNone(self.crawler.get_is_image_addable())

    def test_object_bound_method_deleted(self) -> None:
        # arrange
        obj = _C()
        obj_wr = weak_ref(obj)
        # act
        self.crawler.set_is_image_addable(obj.m_b)
        self.assertEqual(self.crawler.get_is_image_addable(), obj.m_b)
        del obj
        # assert
        self.assertIsNone(obj_wr(), 'obj is intended to be deleted')
        self.assertIsNone(self.crawler.get_is_image_addable())

    def test_function(self) -> None:
        """ Remember: FunctionType is LambdaType """
        # assert
        with self.assertRaises(Exception):
            # currently not supporting `function`. write the test, when writing it is supported
            self.crawler.set_is_image_addable(_f_b)


class CrawlerImageAddedTestCase(unittest.TestCase):
    """
    test if the weakref is working as expected
    """

    def setUp(self) -> None:
        self.crawler = Crawler(MockableImageCrawler(), 1)

    def tearDown(self) -> None:
        del self.crawler

    def test_default(self) -> None:
        # assert
        self.assertIsNone(self.crawler.get_image_added())

    def test_none(self) -> None:
        # arrange
        obj = _C()
        self.crawler.set_image_added(obj.m)
        # act
        self.crawler.set_image_added(None)
        # assert
        self.assertIsNone(self.crawler.get_image_added())

    def test_object_bound_method_stay(self) -> None:
        # arrange
        obj = _C()
        # act
        self.crawler.set_image_added(obj.m)
        # assert
        self.assertEqual(self.crawler.get_image_added(), obj.m)

    def test_object_bound_method_lambda(self) -> None:
        # arrange
        obj = _C()
        self.crawler.set_image_added(obj.m)
        # act
        self.crawler.set_image_added(_C().m)
        # assert
        self.assertIsNone(self.crawler.get_image_added())

    def test_object_bound_method_deleted(self) -> None:
        # arrange
        obj = _C()
        obj_wr = weak_ref(obj)
        # act
        self.crawler.set_image_added(obj.m)
        self.assertEqual(self.crawler.get_image_added(), obj.m)
        del obj
        # assert
        self.assertIsNone(obj_wr(), 'obj is intended to be deleted')
        self.assertIsNone(self.crawler.get_image_added())

    def test_function(self) -> None:
        """ Remember: FunctionType is LambdaType """
        # assert
        with self.assertRaises(Exception):
            # currently not supporting `function`. write the test, when writing it is supported
            self.crawler.set_image_added(_f)


class CrawlerCrawlCase(unittest.TestCase):

    def setUp(self) -> None:
        self.images = ImageCollection({Image(uri='1', source='test'), Image(uri='2', source='test')})
        self.imagecrawler = MockableImageCrawler()
        self.imagecrawler.crawl = MagicMock(return_value=self.images)  # type: ignore
        self.crawler = Crawler(self.imagecrawler, 1)

    def tearDown(self) -> None:
        del self.images
        del self.imagecrawler
        del self.crawler

    def test_crawl_default(self) -> None:
        # act
        crawled = self.crawler.crawl()
        # assert
        self.assertSetEqual(self.images, self.crawler.images)
        self.assertEqual(len(self.images), crawled)

    def test_crawl_is_addable_true(self) -> None:
        called_is_image_addable_with = ImageCollection()

        def on_is_addable_true(image: Image) -> bool:
            called_is_image_addable_with.add(image)
            return False

        self.crawler.get_is_image_addable = MagicMock(return_value=on_is_addable_true)  # type: ignore
        # act
        crawled = self.crawler.crawl()
        # assert
        self.assertSetEqual(self.images, called_is_image_addable_with)
        self.assertEqual(len(self.images), crawled)

    def test_crawl_is_addable_false(self) -> None:
        # arrange
        called_is_image_addable_with = ImageCollection()

        def on_is_addable_false(image: Image) -> bool:
            called_is_image_addable_with.add(image)
            return False

        self.crawler.get_is_image_addable = MagicMock(return_value=on_is_addable_false)  # type: ignore
        # act
        crawled = self.crawler.crawl()
        # assert
        self.assertSetEqual(self.images, called_is_image_addable_with)
        self.assertEqual(len(self.images), crawled)

    def test_crawl_image_added_called(self) -> None:
        # arrange
        called_image_added_with = ImageCollection()

        def on_get_image_added(image: Image) -> bool:
            called_image_added_with.add(image)
            return False

        self.crawler.get_image_added = MagicMock(return_value=on_get_image_added)  # type: ignore
        # act
        crawled = self.crawler.crawl()
        # assert
        self.assertSetEqual(self.images, called_image_added_with)
        self.assertEqual(len(self.images), crawled)


class ServerRefillTest(unittest.TestCase):

    def test_fill_up_to(self) -> None:
        self.skipTest("write the test - use Random3Crawler")

    def test_refill_null_crawler(self) -> None:
        self.skipTest("write the test - use NullCrawler")



import unittest
from typing import List, Optional
from unittest.mock import MagicMock

from ddt import data as ddt_data, ddt, unpack as ddt_unpack  # type: ignore

from nichtparasoup.core import Crawler, CrawlerCollection, _CrawlerWeight

from .mockable_imagecrawler import MockableImageCrawler


@ddt
class CrawlersTest(unittest.TestCase):

    # TODO: remove ddt here, also from setup.py - write proper tests with a reasonable case
    @ddt_data(  # type: ignore
        ([], 0, None), ([], 1, None),
        ([1], 0, 0), ([1], 0.5, 0), ([1], 1, 0),
        ([1, 2], 0, 0), ([0.7, 2], 0.5, 0), ([1, 2], 1.001, 1), ([1, 2], 3, 1),
    )
    @ddt_unpack  # type: ignore
    def test_get_random(self, weights: List[_CrawlerWeight],
                        weight_rnd: _CrawlerWeight, expected_choice: Optional[int]
                        ) -> None:
        # arrange
        crawlers = CrawlerCollection()
        crawlers._random_weight = MagicMock(return_value=weight_rnd)  # type: ignore
        for weight in weights:
            crawlers.append(Crawler(MockableImageCrawler(), weight))

        # act
        random_crawler = crawlers.get_random()

        # assert
        if expected_choice is None:
            self.assertIsNone(random_crawler, "did not expect choice")
        else:
            self.assertIs(random_crawler, crawlers[expected_choice], "did expect different coice")



import unittest


class ArgParserTest(unittest.TestCase):

    def setUp(self) -> None:
        from nichtparasoup.cmdline.argparse import parser
        self.parser = parser

    def tearDown(self) -> None:
        del self.parser

    def test_parser(self) -> None:
        self.skipTest("TODO: write the test")  # TODO: write the test



__all__ = ['_logger', '_log', '_message', '_message_exception', '_confirm']
"""
yes, everything is underscored.
its internal foo that is not for public use.
"""

import logging
import sys
from typing import Any, Optional, TextIO

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

try:
    from termcolor import colored
except ImportError:
    colored = None  # type: ignore

_logger = logging.getLogger('nichtparasoup')

_LOG_LEVEL = Literal['debug', 'info', 'warning', 'error', 'critical', 'log', 'exception']


def _log(level: _LOG_LEVEL, message: str, *args: Any, **kwargs: Any) -> None:
    if not logging.root.handlers and _logger.level == logging.NOTSET:
        _logger.setLevel(logging.INFO)
        _logger.addHandler(logging.StreamHandler())
    getattr(_logger, level)(message.rstrip(), *args, **kwargs)


def _message(message: str, color: Optional[str] = None, file: Optional[TextIO] = None) -> None:
    from sys import stdout
    newline = '\r\n'
    if not file:
        file = stdout
    if color and colored:
        message = colored(message, color=color)
    file.write('{}{}'.format(message.rstrip(), newline))


def _message_exception(exception: BaseException, file: Optional[TextIO] = None) -> None:
    if not file:
        from sys import stderr
        file = stderr
    exception_name = type(exception).__name__
    if colored:
        color = 'yellow' if isinstance(exception, Warning) else 'red'
        exception_name = colored(exception_name, color)
    _message('{}: {}'.format(exception_name, exception), file=file)


def _confirm(prompt: str, default: bool = False) -> Optional[bool]:
    rv = {
        'y': True,
        'yes': True,
        '': default,
        'n': False,
        'no': False,
    }
    options = 'Y/n' if default else 'y/N'
    try:
        value = input('{!s} [{}]: '.format(prompt, options)).lower().strip()
        return rv[value]
    except (KeyboardInterrupt, EOFError, KeyError):
        return None



__all__ = ["MockableImageCrawler", "YetAnotherImageCrawler"]

from typing import Any, Dict

from nichtparasoup.core.image import ImageCollection
from nichtparasoup.core.imagecrawler import BaseImageCrawler, ImageCrawlerConfig, ImageCrawlerInfo


class MockableImageCrawler(BaseImageCrawler):
    """ imagecrawler that does nothing. use it for mocking ... """

    @classmethod
    def info(cls) -> ImageCrawlerInfo:
        return ImageCrawlerInfo(
            description='a mock',
            config=dict(),
        )

    @classmethod
    def check_config(cls, config: Dict[Any, Any]) -> ImageCrawlerConfig:
        return ImageCrawlerConfig(config)

    def _reset(self) -> None:
        pass

    def _crawl(self) -> ImageCollection:
        return ImageCollection()


class YetAnotherImageCrawler(MockableImageCrawler):
    """ another implementation, to see if type matters """
    pass



import unittest
from os.path import dirname, join as path_join, realpath

from nichtparasoup.config import parse_yaml_file


class ConfigParserDefaultsTest(unittest.TestCase):

    def test_set_optional_loglevel(self) -> None:
        # arrange
        file = realpath(path_join(dirname(__file__), 'configs', 'positive', 'logging_level_missing.yaml'))
        # act
        config = parse_yaml_file(file)
        # assert
        self.assertEqual(config['logging']['level'], 'INFO')

    def test_set_optional_weight(self) -> None:
        # arrange
        file = realpath(path_join(dirname(__file__), 'configs', 'positive', 'crawler_weight_missing.yaml'))
        # act
        config = parse_yaml_file(file)
        # assert
        self.assertEqual(len(config["crawlers"]), 1)
        self.assertEqual(config["crawlers"][0]["weight"], 1)

    def test_set_optional_config(self) -> None:
        # arrange
        file = realpath(path_join(dirname(__file__), 'configs', 'positive', 'crawler_config_missing.yaml'))
        # act
        config = parse_yaml_file(file)
        # assert
        self.assertEqual(len(config["crawlers"]), 1)
        self.assertDictEqual(config["crawlers"][0]["config"], dict())



#!/usr/bin/env python3

from setuptools import find_packages, setup

setup(
    name='nichtparasoup-imagecrawler-placeholders',
    version='0.0.1',
    description='PlaceholderImages for UI tests',
    python_requires='>=3.5',
    packages=find_packages(exclude=['tests', '*.tests', '*.tests.*', 'tests.*', 'examples']),
    entry_points=dict(
        nichtparasoup_imagecrawler=[
            'DummyImage = nichtparasoup_placeholders:DummyImage',
        ],
    ),
    install_requires=[
        'nichtparasoup',
        # This example requires `nichtparasoup>=2.3`
        # You SHOULD write the min-version dependency in your real-world.
        # This was not done here, since its an in-dev example.
    ],
    extras_require=dict(
        testing=[
            'mypy',
            'pytest>=5,<5.3.4',  # 5.3.4 has a bug: https://github.com/pytest-dev/pytest/issues/6517
            'coverage',
            'ddt',
        ]
    ),
)



import unittest
from typing import Any, Dict

from nichtparasoup.core import NPCore
from nichtparasoup.core.server import Server, ServerStatus

from .mockable_imagecrawler import MockableImageCrawler


class ServerStatusStableTest(unittest.TestCase):

    def setUp(self) -> None:
        core = NPCore()
        self.imagecrawlers = ((MockableImageCrawler(t=1), 1), (MockableImageCrawler(t=2), 1))
        for (imagecrawler, weight) in self.imagecrawlers:
            core.add_imagecrawler(imagecrawler, weight)
        self.server = Server(core)

    def tearDown(self) -> None:
        del self.server

    def test_server(self) -> None:
        # act
        status = ServerStatus.server(self.server)
        # assert
        self.assertIsInstance(status, dict)
        self.assertIsInstance(status.get("version"), str)
        self.assertIsInstance(status.get("uptime"), int)
        self.assertIsInstance(status.get("reset"), dict)
        status_reser = status["reset"]  # type: Dict[Any, Any]
        self.assertIsInstance(status_reser.get("since"), int)
        self.assertIsInstance(status_reser.get("count"), int)
        self.assertIsInstance(status.get("images"), dict)
        status_images = status["images"]  # type: Dict[Any, Any]
        self.assertIsInstance(status_images.get("served"), int)
        self.assertIsInstance(status_images.get("crawled"), int)

    def test_blacklist(self) -> None:
        # act
        status = ServerStatus.blacklist(self.server)
        # assert
        self.assertIsInstance(status, dict)
        self.assertIsInstance(status.get("len"), int)
        self.assertIsInstance(status.get("size"), int)

    def test_crawlers(self) -> None:
        # act
        status = ServerStatus.crawlers(self.server)
        # assert
        self.assertIsInstance(status, dict)
        self.assertEqual(len(self.imagecrawlers), len(status))
        for crawler in self.server.core.crawlers:
            crawler_id = id(crawler)
            self.assertIsInstance(status.get(crawler_id), dict)
            crawler_status = status[crawler_id]  # type: Dict[Any, Any]
            self.assertIsInstance(crawler_status, dict)
            self.assertIsInstance(crawler_status.get("type"), str)
            self.assertIsInstance(crawler_status.get("weight"), (int, float))
            self.assertIsInstance(crawler_status.get("config"), dict)
            self.assertIsInstance(crawler_status.get("images"), dict)
            crawler_status_images = crawler_status["images"]  # type: Dict[Any, Any]
            self.assertIsInstance(crawler_status_images.get("len"), int)
            self.assertIsInstance(crawler_status_images.get("size"), int)



import pytest  # type: ignore

from nichtparasoup.config import _DEFAULTS_FILE
from nichtparasoup.testing.config import ConfigFileTest


class DefaultConfigFileTest(ConfigFileTest):

    @pytest.mark.online  # type: ignore  # probing goes over the internet
    @pytest.mark.slow  # type: ignore  # probing might be slow
    def test_default(self) -> None:
        self.validate(_DEFAULTS_FILE)
        self.probe(_DEFAULTS_FILE)



import logging
from typing import Any, List, Optional

from nichtparasoup._internals import _message, _message_exception


def _logging_init(level: int) -> None:  # pragma: no cover
    if not logging.root.handlers:
        logging.root.setLevel(level)
        logging.root.addHandler(logging.StreamHandler())


class Commands(object):

    @staticmethod
    def run(config_file: Optional[str] = None) -> int:
        from os.path import abspath
        from nichtparasoup.config import get_config, get_imagecrawler
        from nichtparasoup.core import NPCore
        from nichtparasoup.core.server import Server as ImageServer
        from nichtparasoup.webserver import WebServer
        try:
            config = get_config(abspath(config_file) if config_file else None)
            _logging_init(getattr(logging, config['logging']['level']))
            imageserver = ImageServer(NPCore(), **config['imageserver'])
            for crawler_config in config['crawlers']:
                imagecrawler = get_imagecrawler(crawler_config)
                if not imageserver.core.has_imagecrawler(imagecrawler):
                    imageserver.core.add_imagecrawler(imagecrawler, crawler_config['weight'])
            webserver = WebServer(imageserver, **config['webserver'])
            webserver.run()
            return 0
        except Exception as e:
            _message_exception(e)
            return 1

    @classmethod
    def config(cls, action: str, config_file: str) -> int:
        return dict(
            check=cls.config_check_file,
            dump=cls.config_dump_file,
        )[action](config_file)

    @staticmethod
    def config_dump_file(config_file: str) -> int:
        from os.path import abspath, isfile
        from nichtparasoup.config import dump_defaults
        config_file = abspath(config_file)
        if isfile(config_file):
            from nichtparasoup._internals import _confirm
            overwrite = _confirm('File already exists, overwrite?')
            if overwrite is not True:
                _message('Abort.')
                return 1
        try:
            dump_defaults(config_file)
            return 0
        except Exception as e:
            _message_exception(e)
            return 255

    @staticmethod
    def config_check_file(config_file: str) -> int:
        from os.path import abspath
        from nichtparasoup.testing.config import ConfigFileTest
        config_file = abspath(config_file)
        config_test = ConfigFileTest()
        try:
            config_test.validate(config_file)
            config_test.probe(config_file)
        except Exception as e:
            _message_exception(e)
            return 255
        return 0

    @classmethod
    def info(cls, **actions: Any) -> int:
        active_actions = dict((k, v) for k, v in actions.items() if v)
        if len(active_actions) != 1:
            _message_exception(ValueError('exactly one action required'))
            return 255
        action, action_value = active_actions.popitem()
        return dict(
            version=cls.info_version,
            imagecrawler_list=cls.info_imagecrawler_list,
            imagecrawler_desc=cls.info_imagecrawler_desc,
        )[action](action_value)

    @staticmethod
    def info_version(_: Any) -> int:
        from nichtparasoup import VERSION
        _message(VERSION)
        return 0

    @staticmethod
    def info_imagecrawler_list(_: Any) -> int:
        from nichtparasoup.imagecrawler import get_imagecrawlers
        imagecrawlers = get_imagecrawlers().names()
        if not imagecrawlers:
            _message_exception(Warning('no ImageCrawler found'))
        else:
            _message("\r\n".join(sorted(imagecrawlers)))
        return 0

    @staticmethod
    def info_imagecrawler_desc(imagecrawler: str) -> int:
        from nichtparasoup._internals import _log
        from nichtparasoup.core.server import type_module_name_str
        from nichtparasoup.imagecrawler import get_imagecrawlers
        imagecrawler_class = get_imagecrawlers().get_class(imagecrawler)
        if not imagecrawler_class:
            _message_exception(ValueError('unknown ImageCrawler {!r}'.format(imagecrawler)))
            return 1
        info = []
        info_linebreak = '\r\n'
        imagecrawler_info = imagecrawler_class.info()
        info.append(imagecrawler_info.description)
        if imagecrawler_info.long_description:
            info.append(imagecrawler_info.long_description)
        if imagecrawler_info.config:
            info_bull = info_linebreak + ' * '
            mlen = max(len(k) for k in imagecrawler_info.config.keys())
            info.append('Config: ' + info_bull + info_bull.join([
                '{key:{mlen}}: {desc}'.format(mlen=mlen, key=key, desc=desc)
                for key, desc in imagecrawler_info.config.items()]))
        _message((info_linebreak * 2).join(info))
        _log('debug', info_linebreak.join(
            [
                info_linebreak,
                'DEBUG INFO',
                'Icon : {!r}'.format(imagecrawler_info.icon_url),
                'Class: {!r}'.format(type_module_name_str(imagecrawler_class)),
            ]))
        return 0


def main(args: Optional[List[str]] = None) -> int:
    from nichtparasoup.cmdline.argparse import parser as argparser
    options = dict(argparser.parse_args(args=args).__dict__)
    if options.pop('debug', False):
        _logging_init(logging.DEBUG)
        _message('DEBUG ENABLED :)', 'cyan')
    command = options.pop('command')
    return getattr(Commands, command)(**options)  # type: ignore



__all__ = ["VERSION"]

try:
    # file generated by setuptools_scm
    from nichtparasoup.__version__ import version
    VERSION = str(version or 'UNKNOWN.falsy')
except ImportError:  # pragma: no cover
    VERSION = 'UNKNOWN.import'



import unittest

from .mockable_imagecrawler import MockableImageCrawler, YetAnotherImageCrawler


class BaseImageCrawlerEqualTest(unittest.TestCase):

    def test_equal(self) -> None:
        # arrange
        c01 = MockableImageCrawler()
        c02 = MockableImageCrawler()
        c11 = MockableImageCrawler(foo='bar', baz=1)
        c12 = MockableImageCrawler(foo='bar', baz=1)
        # assert
        self.assertEqual(c01, c01)
        self.assertEqual(c02, c02)
        self.assertEqual(c01, c02)
        self.assertEqual(c11, c11)
        self.assertEqual(c12, c12)
        self.assertEqual(c11, c12)

    def test_unequal(self) -> None:
        # arrange
        c1 = MockableImageCrawler()
        c2 = YetAnotherImageCrawler()
        c3 = MockableImageCrawler(foo='bar', baz=1)
        c4 = YetAnotherImageCrawler(foo='b4r', baz=1)
        c5 = MockableImageCrawler(baz=1)
        # assert
        self.assertNotEqual(c1, c2)
        self.assertNotEqual(c1, c3)
        self.assertNotEqual(c1, c4)
        self.assertNotEqual(c1, c5)
        self.assertNotEqual(c2, c3)
        self.assertNotEqual(c2, c4)
        self.assertNotEqual(c2, c5)
        self.assertNotEqual(c3, c4)
        self.assertNotEqual(c3, c5)
        self.assertNotEqual(c4, c5)


class BaseImageCrawlerResetTest(unittest.TestCase):

    def test_reset_set(self) -> None:
        # arrange
        c = MockableImageCrawler()
        c._reset_before_next_crawl = False
        # act
        c.reset()
        # assert
        self.assertTrue(c._reset_before_next_crawl)

    def test_reset_released(self) -> None:
        # arrange
        c = MockableImageCrawler()
        c._reset_before_next_crawl = True
        # act
        c.crawl()
        # assert
        self.assertFalse(c._reset_before_next_crawl)


class BaseImageCrawlerGetConfigTest(unittest.TestCase):

    def test_public(self) -> None:
        # arrange
        c = MockableImageCrawler(foo='bar')
        # act
        config = c.get_config()
        # assert
        self.assertDictEqual(dict(foo='bar'), config)

    def test_empty_key(self) -> None:
        # arrange
        c = MockableImageCrawler(**{'': 'bar'})
        # act
        config = c.get_config()
        # assert
        self.assertDictEqual({'': 'bar'}, config)

    def test_protected(self) -> None:
        # arrange
        c = MockableImageCrawler(_foo='bar')
        # act
        config = c.get_config()
        # assert
        self.assertDictEqual(dict(), config)

    def test_private(self) -> None:
        # arrange
        c = MockableImageCrawler(__foo='bar')
        # act
        config = c.get_config()
        # assert
        self.assertDictEqual(dict(), config)



__all__ = ["InstagramHashtag", "InstagramProfile"]

import sys
from abc import ABC, abstractmethod
from json import dumps as json_encode, loads as json_loads
from re import compile as re_compile
from threading import Lock
from typing import Any, Dict, Optional, Pattern, Set
from urllib.parse import quote_plus as url_quote, urlencode, urljoin

from nichtparasoup.imagecrawler import (
    BaseImageCrawler, Image, ImageCollection, ImageCrawlerConfig, ImageCrawlerInfo, RemoteFetcher,
)

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

"""

INFO
====

Instagra provides an API. But it requires a access token, which requires an account ... don't have that.
Intagram provides a JSON based web endpoint, that is used for lazy loading on the website. Let's use it.

WEB ENDPOINT
------------

The web end point takes some GET data: data is identified by `query_hash` and `<misc>`
From reverse engineering this seams to be valid information.
* query_hash - Like an access token. generated on first contact.
               This one is kind of consistent for certain network topology constraints.
               Important: hash is bound to a purpose: tags, profile, meta, user_info, ...
* variables  - JSON encoded map of:
    * first  - Number how many results are expected.
               This seams to be a vague idea. mostly you get more than expected.
    * after  - A cursor in a sliding window over the data
    * <misc> - Identifier of condition/clause.
               * 'id' for profiles
               * 'tag_name' for tags
               * etc.

INTERNALS
---------

Since `query_hash` ins unique per purpose, its fetched once per class. so instances do share it.
Finding possible `query_hash` is quite easy my searching through instagram's root pge and included JavaScripts.
If a candidate is correct will be checked, if instagram's response to a request has the correct data format. This is
done by checking a response for a certain format that is unique to its purpose (see `query_hash` description)

The actual response parsing is quite trivial... read the code, you will get it.

CONCLUSION
----------

This implementation seams pretty complex.
If this crawler breaks, it might be a damn good idea to switch to a foreign library
or maybe even use the official API.

"""

INSTAGRAM_URL_ROOT = 'https://www.instagram.com/'

INSTAGRAM_ICON_URL = INSTAGRAM_URL_ROOT + 'static/images/ico/favicon-192.png/68d99ba29cc8.png'

_InstagramQueryHashFinder_ContainerType = Literal['tag', 'profile']


class InstagramQueryHashFinder(object):
    __CONTAINER_PATH_RE = {
        'tag': r'/static/bundles/metro/TagPageContainer\.js/.+?\.js',
        'profile': r'/static/bundles/metro/ProfilePageContainer\.js/.+?\.js',
    }

    __QUERY_HASH_RE = r'queryId:"(.+?)"'

    def __init__(self, container_type: _InstagramQueryHashFinder_ContainerType) -> None:  # pragma: no cover
        self._container_re = re_compile(self.__class__.__CONTAINER_PATH_RE[container_type])
        self._query_hash_re = re_compile(self.__class__.__QUERY_HASH_RE)
        self._remote_fetcher = RemoteFetcher()

    def find_hashes(self) -> Set[str]:
        return self._get_from_container(self._container_re)

    def _get_from_container(self, re_container_path: Pattern[str]) -> Set[str]:
        page_doc, page_uri = self._remote_fetcher.get_string(INSTAGRAM_URL_ROOT)
        container_paths = re_container_path.search(page_doc)
        if container_paths:
            container_path = urljoin(page_uri, container_paths.group(0))
            return self._get_from_remote_js(container_path)
        raise InstagramError('container not found')

    def _get_from_remote_js(self, js_uri: str) -> Set[str]:
        try:
            js_src, _ = self._remote_fetcher.get_string(js_uri)
        except Exception:
            return set()
        return self._get_from_js(js_src)

    def _get_from_js(self, js: str) -> Set[str]:
        return set(self._query_hash_re.findall(js))


class BaseInstagramCrawler(BaseImageCrawler, ABC):

    def __init__(self, **config: Any) -> None:  # pragma: no cover
        super().__init__(**config)
        self._amount = 10
        self._cursor = None  # type: Optional[str]
        self._remote_fetcher = RemoteFetcher()

    def _reset(self) -> None:
        self._cursor = None

    def _crawl(self) -> ImageCollection:
        images = ImageCollection()
        query_uri = self._get_query_uri(
            self._get_query_hash(), self._amount, self._cursor, **self._get_query_variables())
        response = self._query(query_uri)
        for edge in response['edges']:
            images.update(self.__class__._get_images_from_media_edge_node(edge['node']))
            del edge
        page_info = response['page_info']  # type: Dict[str, Any]
        # don't care if this was the last page ... why not restarting at front when the end is reached?!
        self._cursor = page_info['end_cursor'] if page_info['has_next_page'] else None
        return images

    @classmethod
    def _get_images_from_media_edge_node(cls, node: Dict[str, Any]) -> ImageCollection:
        images = ImageCollection()
        if not node['is_video']:
            source = cls._get_post_url(node['shortcode'])
            images.add(Image(
                uri=node['display_url'],
                source=source,
                dimensions=node.get('dimensions'),
            ))
            for side_edge in node.get('edge_sidecar_to_children', dict(edges=[]))['edges']:
                if not side_edge['node']['is_video']:
                    images.add(Image(
                        uri=side_edge['node']['display_url'],
                        source=source,
                        dimensions=side_edge['node'].get('dimensions'),
                    ))
                del side_edge
        return images

    __URL_POST_TEMPLATE = INSTAGRAM_URL_ROOT + 'p/{}/'

    @classmethod
    def _get_post_url(cls, shortcode: str) -> str:
        return cls.__URL_POST_TEMPLATE.format(url_quote(shortcode))

    def _query(self, uri: str) -> Dict[str, Any]:
        response_string, uri = self._remote_fetcher.get_string(uri)
        # responses may be small in size but are memory hungry when parsing!
        response = json_loads(response_string)
        del response_string
        if response['status'] != 'ok':
            raise InstagramError('response not ok')
        try:
            return self._get_media_from_query_response(response)
        except KeyError as e:
            raise InstagramError('no media') from e

    @classmethod
    @abstractmethod
    def _get_media_from_query_response(cls, response: Dict[str, Any]) -> Dict[str, Any]:  # pragma: no cover
        """
        get the path for media edges in query response

        example implementation:
            return response['data']['<aTYPE>']['edge_<bTYPE>_media']
        """
        raise NotImplementedError()

    __URL_QUERY = INSTAGRAM_URL_ROOT + 'graphql/query/'

    def _get_query_uri(self, query_hash: str, first: int, after: Optional[str], **variables: Any) -> str:
        return self.__class__.__URL_QUERY + '?' + urlencode(dict(
            query_hash=query_hash,
            variables=json_encode(dict(
                first=first,
                after=(after or ""),
                **variables
            ))
        ))

    @abstractmethod
    def _get_queryhashfinder(self) -> InstagramQueryHashFinder:  # pragma: no cover
        """
        get the query hash finder for this type of crawler

        example implementation:
            return InstagramQueryHashFinder('<type>')
        """
        raise NotImplementedError()

    def _check_query_hash(self, query_hash: str) -> bool:
        uri = self._get_query_uri(query_hash, 1, None, **self._get_query_variables())
        try:
            self._query(uri)
            return True
        except Exception:
            return False

    def _find_query_hash(self) -> Optional[str]:
        query_hashes = self._get_queryhashfinder().find_hashes()
        for query_hash in query_hashes:
            if self._check_query_hash(query_hash):
                return query_hash
        return None

    _QUERY_HASH_LOCK = Lock()  # global lock. may be overwritten in subclass
    _query_hash = None  # type: Optional[str]

    def _get_query_hash(self) -> str:
        cls = self.__class__
        # same class = same query_hash ... so lock and search ... others may use the same hash later
        with cls._QUERY_HASH_LOCK:
            if not cls._query_hash:
                query_hash = self._find_query_hash()
                if not query_hash:
                    raise InstagramError('did not find query hash')
                cls._query_hash = query_hash
        return cls._query_hash

    @abstractmethod
    def _get_query_variables(self) -> Dict[str, Any]:  # pragma: no cover
        """
        return the variables that are required for a query url

        example implementation:
            return dict(foo='bar')
        """
        raise NotImplementedError()


class InstagramHashtag(BaseInstagramCrawler):
    _QUERY_HASH_LOCK = Lock()  # all instances of HashtagCrawler share the same query_hash - and this lock

    @classmethod
    def info(cls) -> ImageCrawlerInfo:
        return ImageCrawlerInfo(
            description='A Crawler for hash tag posts of https://www.instagram.com/',
            config=dict(
                tag_name='the HashTag on Instagram',
            ),
            icon_url=INSTAGRAM_ICON_URL,
        )

    @classmethod
    def check_config(cls, config: Dict[Any, Any]) -> ImageCrawlerConfig:
        tag_name = config["tag_name"]
        if type(tag_name) is not str:
            raise TypeError("tag_name {!r} is not str".format(tag_name))
        if 0 == len(tag_name):
            raise ValueError("tag_name {!r} is empty".format(tag_name))
        return ImageCrawlerConfig(
            tag_name=tag_name.lower(),
        )

    def _get_queryhashfinder(self) -> InstagramQueryHashFinder:
        return InstagramQueryHashFinder('tag')

    def _get_query_variables(self) -> Dict[str, Any]:
        return dict(tag_name=self._config['tag_name'])

    @classmethod
    def _get_media_from_query_response(cls, response: Dict[str, Any]) -> Dict[str, Any]:
        return response['data']['hashtag']['edge_hashtag_to_media']  # type: ignore


class InstagramProfile(BaseInstagramCrawler):
    _QUERY_HASH_LOCK = Lock()  # all instances of ProfileCrawler share the same query_hash - and this lock

    def __init__(self, **config: Any) -> None:  # pragma: no cover
        super().__init__(**config)
        self.__profile_id = None  # type: Optional[str]
        self.__profile_id_lock = Lock()

    @classmethod
    def info(cls) -> ImageCrawlerInfo:
        from nichtparasoup import VERSION
        return ImageCrawlerInfo(
            description='A Crawler for profile pages of https://www.instagram.com/',
            config=dict(
                user_name='the UserName on Instagram',
            ),
            version=VERSION,
            icon_url=INSTAGRAM_ICON_URL,
        )

    @classmethod
    def check_config(cls, config: Dict[Any, Any]) -> ImageCrawlerConfig:
        user_name = config["user_name"]
        if type(user_name) is not str:
            raise TypeError("user_name {!r} is not str".format(user_name))
        if 0 == len(user_name):
            raise ValueError("user_name {!r} is empty".format(user_name))
        return ImageCrawlerConfig(
            user_name=user_name.lower(),
        )

    def _get_queryhashfinder(self) -> InstagramQueryHashFinder:
        return InstagramQueryHashFinder('profile')

    @classmethod
    def _get_media_from_query_response(cls, response: Dict[str, Any]) -> Dict[str, Any]:
        return response['data']['user']['edge_owner_to_timeline_media']  # type: ignore

    def _get_query_variables(self) -> Dict[str, Any]:
        return dict(id=self._get_profile_id())

    def _fetch_profile(self) -> Dict[str, Any]:
        # this is much easier than parsing `window._sharedData` from the website - let's hope it is stable again
        profile_string, _ = self._remote_fetcher.get_string(self._get_profile_url() + '?__a=1')
        return json_loads(profile_string)  # type: ignore

    def _fetch_profile_id(self) -> str:
        profile = self._fetch_profile()
        try:
            return profile['graphql']['user']['id']  # type: ignore
        except KeyError as e:
            raise InstagramError('profile_id not found') from e

    def _get_profile_id(self) -> str:
        with self.__profile_id_lock:
            if self.__profile_id is None:
                self.__profile_id = self._fetch_profile_id()
        return self.__profile_id

    __PROFILE_URL_TEMPLATE = INSTAGRAM_URL_ROOT + '{}/'

    def _get_profile_url(self) -> str:
        return self.__class__.__PROFILE_URL_TEMPLATE.format(url_quote(self._config['user_name']))


class InstagramError(Exception):
    pass



import unittest
from collections import OrderedDict
from json import dumps as json_dumps, loads as json_loads
from os.path import dirname, join as path_join
from typing import Any, Dict, Type
from urllib.parse import parse_qs, urlencode, urlparse, urlunparse

from nichtparasoup.imagecrawler import BaseImageCrawler, Image, ImageCollection
from nichtparasoup.imagecrawler.instagram import (
    BaseInstagramCrawler, InstagramHashtag, InstagramProfile, InstagramQueryHashFinder,
)
from nichtparasoup.testing.imagecrawler import FileFetcher, ImageCrawlerLoaderTest


class _InstagramFileFetcher(FileFetcher):

    @classmethod
    def _uri_sort_query(cls, uri: str) -> str:
        scheme, netloc, path, params, query, fragment = urlparse(uri)
        if query == '':
            query_sorted = query
        else:
            query_dict = parse_qs(query, keep_blank_values=True)
            if 'variables' in query_dict:
                variables = query_dict['variables'][0]
                variables_dict = json_loads(variables)
                variables_dict_sorted = OrderedDict((k, variables_dict[k]) for k in sorted(variables_dict))
                query_dict['variables'][0] = json_dumps(variables_dict_sorted)
                query_sorted = urlencode(query_dict, doseq=True)
            else:
                query_sorted = query
        uri_sorted = urlunparse((scheme, netloc, path, params, query_sorted, fragment))
        return super()._uri_sort_query(uri_sorted)


_FILE_FETCHER = _InstagramFileFetcher({  # relative to './testdata_instagram'
    '/': 'index.html',
    '/static/bundles/metro/ProfilePageContainer.js/e243abb1f92c.js': 'e243abb1f92c.js',
    '/static/bundles/metro/TagPageContainer.js/1bad9348735e.js': '1bad9348735e.js',
    '/graphql/query/?query_hash=f0986789a5c5d17c2400faebf16efd0d&'
    'variables=%7B%22first%22%3A+1%2C+%22after%22%3A+%22%22%2C+%22tag_name%22%3A+%22foo%22%7D':
        'query_hash=f0986789a5c5d17c2400faebf16efd0d&variables={first-1,after,tag_name-foo}',
    '/graphql/query/?query_hash=174a5243287c5f3a7de741089750ab3b&'
    'variables=%7B%22first%22%3A+1%2C+%22after%22%3A+%22%22%2C+%22tag_name%22%3A+%22foo%22%7D':
        'query_hash=174a5243287c5f3a7de741089750ab3b&variables={first-1,after,tag_name-foo}',
    '/graphql/query/?query_hash=ff260833edf142911047af6024eb634a&'
    'variables=%7B%22first%22%3A+1%2C+%22after%22%3A+%22%22%2C+%22tag_name%22%3A+%22foo%22%7D':
        'query_hash=ff260833edf142911047af6024eb634a&variables={first-1,after,tag_name-foo}',
    '/graphql/query/?query_hash=174a5243287c5f3a7de741089750ab3b&'
    'variables=%7B%22first%22%3A+5%2C+%22after%22%3A+%22%22%2C+%22tag_name%22%3A+%22foo%22%7D':
        'query_hash=174a5243287c5f3a7de741089750ab3b&variables={first-5,after,tag_name-foo}',
    '/natgeo/?__a=1': 'natgeo.__a=1',
    '/graphql/query/?query_hash=2c5d4d8b70cad329c4a6ebe3abb6eedd&'
    'variables=%7B%22first%22%3A+1%2C+%22after%22%3A+%22%22%2C+%22id%22%3A+%22787132%22%7D':
        'query_hash=2c5d4d8b70cad329c4a6ebe3abb6eedd&variables={first-1,after,id-787132}',
    '/graphql/query/?query_hash=ff260833edf142911047af6024eb634a&'
    'variables=%7B%22first%22%3A+1%2C+%22after%22%3A+%22%22%2C+%22id%22%3A+%22787132%22%7D':
        'query_hash=ff260833edf142911047af6024eb634a&variables={first-1,after,id-787132}',
    '/graphql/query/?query_hash=f0986789a5c5d17c2400faebf16efd0d&'
    'variables=%7B%22first%22%3A+1%2C+%22after%22%3A+%22%22%2C+%22id%22%3A+%22787132%22%7D':
        'query_hash=f0986789a5c5d17c2400faebf16efd0d&variables={first-1,after,id-787132}',
    '/graphql/query/?query_hash=8c86fed24fa03a8a2eea2a70a80c7b6b&'
    'variables=%7B%22first%22%3A+1%2C+%22after%22%3A+%22%22%2C+%22id%22%3A+%22787132%22%7D':
        'query_hash=8c86fed24fa03a8a2eea2a70a80c7b6b&variables={first-1,after,id-787132}',
    '/graphql/query/?query_hash=2c5d4d8b70cad329c4a6ebe3abb6eedd&'
    'variables=%7B%22first%22%3A+5%2C+%22after%22%3A+%22%22%2C+%22id%22%3A+%22787132%22%7D':
        'query_hash=2c5d4d8b70cad329c4a6ebe3abb6eedd&variables={first-5,after,id-787132}',
}, base_dir=path_join(dirname(__file__), 'testdata_instagram'))

_QUERYHASHES_EXPECTED_TAG = {'f0986789a5c5d17c2400faebf16efd0d',
                             'ff260833edf142911047af6024eb634a',
                             '174a5243287c5f3a7de741089750ab3b'}

_QUERYHASHES_EXPECTED_PROFILE = {'f0986789a5c5d17c2400faebf16efd0d',
                                 'ff260833edf142911047af6024eb634a',
                                 '2c5d4d8b70cad329c4a6ebe3abb6eedd',
                                 '8c86fed24fa03a8a2eea2a70a80c7b6b'}


class InstagramQueryHashFinderTest(unittest.TestCase):

    def test_tag_get_from_container(self) -> None:
        # arrange
        finder = InstagramQueryHashFinder('tag')
        finder._remote_fetcher = _FILE_FETCHER
        # act
        hashes = finder.find_hashes()
        # assert
        self.assertSetEqual(hashes, _QUERYHASHES_EXPECTED_TAG)

    def test_profile_get_from_container(self) -> None:
        # arrange
        finder = InstagramQueryHashFinder('profile')
        finder._remote_fetcher = _FILE_FETCHER
        # act
        hashes = finder.find_hashes()
        # assert
        self.assertSetEqual(hashes, _QUERYHASHES_EXPECTED_PROFILE)


class BaseInstagramCrawlerImagesFromMediaEdgeNodeTest(unittest.TestCase):

    def __test(self, node: Dict[str, Any], images_expected: ImageCollection) -> None:
        # act
        images = BaseInstagramCrawler._get_images_from_media_edge_node(node)
        # assert
        self.assertSetEqual(images, images_expected)

    def test_video(self) -> None:
        # arrange
        node = dict(
            is_video=True,
            display_url='display_url',
            shortcode='shortcode',
        )
        images_expected = ImageCollection()
        # act & assert
        self.__test(node, images_expected)

    def test_image(self) -> None:
        # arrange
        node = dict(
            is_video=False,
            display_url='display_url',
            shortcode='shortcode',
        )
        images_expected = ImageCollection()
        images_expected.add(Image(
            uri='display_url',
            source='https://instagram.com/p/shortcode/',
        ))
        # act & assert
        self.__test(node, images_expected)

    def test_image_sidecar_video(self) -> None:
        # arrange
        node = dict(
            is_video=False,
            display_url='display_url',
            shortcode='shortcode',
            edge_sidecar_to_children=dict(
                edges=[
                    dict(
                        node=dict(
                            is_video=True,
                            display_url='side_display_url',
                            shortcode='side_shortcode',
                        )
                    )
                ]
            ),
        )
        images_expected = ImageCollection()
        images_expected.add(Image(
            uri='display_url',
            source='https://instagram.com/p/shortcode/',
        ))
        # act & assert
        self.__test(node, images_expected)

    def test_image_sidecar_image(self) -> None:
        # arrange
        node = dict(
            is_video=False,
            display_url='display_url',
            shortcode='shortcode',
            edge_sidecar_to_children=dict(
                edges=[
                    dict(
                        node=dict(
                            is_video=False,
                            display_url='side_display_url',
                            shortcode='side_shortcode',
                        )
                    )
                ]
            ),
        )
        images_expected = ImageCollection()
        images_expected.add(Image(
            uri='display_url',
            source='https://instagram.com/p/shortcode/',
        ))
        images_expected.add(Image(
            uri='side_display_url',
            source='https://instagram.com/p/side_shortcode/',
        ))
        # act & assert
        self.__test(node, images_expected)


class InstagramHashtagTest(unittest.TestCase):
    _QUERY_HASH = '174a5243287c5f3a7de741089750ab3b'

    def setUp(self) -> None:
        InstagramHashtag._query_hash = None
        self.crawler = InstagramHashtag(tag_name='foo')
        self.crawler._remote_fetcher = _FILE_FETCHER
        self.crawler._amount = 5
        self.crawler._get_queryhashfinder = self._get_queryhashfinder  # type: ignore

    def tearDown(self) -> None:
        del self.crawler
        InstagramHashtag._query_hash = None

    def _get_queryhashfinder(self) -> InstagramQueryHashFinder:
        finder = InstagramQueryHashFinder('tag')
        finder.find_hashes = lambda: _QUERYHASHES_EXPECTED_TAG  # type: ignore
        return finder

    def test__get_query_hash(self) -> None:
        # act
        queryhash = self.crawler._get_query_hash()
        # assert
        self.assertEqual(queryhash, self.__class__._QUERY_HASH)

    def test_reset(self) -> None:
        # arrange
        self.crawler._cursor = 'foo'
        # act
        self.crawler._reset()
        # assert
        self.assertIsNone(self.crawler._cursor)

    def test__crawl(self) -> None:
        # arrange
        self.crawler._get_query_hash = lambda: self.__class__._QUERY_HASH  # type: ignore
        expected_images = ImageCollection()
        expected_cursor = 'QVFDdV9PUXYxc0hjcU9TYUI5dWZZWmNsOGdSaUsxcU9oUHg5endkc2hiUnV' \
                          'CZVVDZWFUM2QzdlVvSnN0Z053Q2ItQkxvSGRObm1hdlR5X3dDZ1JKMWduRg=='

        expected_images.add(Image(
            source='https://www.instagram.com/p/B5F7QoklRxu/',
            uri='https://scontent-frx5-1.cdninstagram.com/v/t51.2885-15/e35/75397747_'
                '403992793822822_8324298994393298267_n.jpg?_nc_ht=scontent-frx5-1.cdninstagram.com'
                '&_nc_cat=110&oh=f3e9d48ba846ef1d7371e450ab8099e7&oe=5E567BF8'
        ))
        expected_images.add(Image(
            source='https://www.instagram.com/p/B5F68piFIAs/',
            uri='https://scontent-frx5-1.cdninstagram.com/v/t51.2885-15/e35/72281363_241905656781018_'
                '6014571893708900461_n.jpg?_nc_ht=scontent-frx5-1.cdninstagram.com'
                '&_nc_cat=106&oh=8b66604749764befc12a2565a2485649&oe=5E6C6C16'
        ))
        expected_images.add(Image(
            source='https://www.instagram.com/p/B5F4v16HTM5/',
            uri='https://scontent-frx5-1.cdninstagram.com/v/t51.2885-15/e35/70998485_152987982771030_'
                '3346050114145538962_n.jpg?_nc_ht=scontent-frx5-1.cdninstagram.com'
                '&_nc_cat=106&oh=5cf81cd1f81acf8565656bd4af311717&oe=5E6FC7D5'
        ))
        expected_images.add(Image(
            source='https://www.instagram.com/p/B5F4qHBHNrD/',
            uri='https://scontent-frx5-1.cdninstagram.com/v/t51.2885-15/e35/p1080x1080/74862751_'
                '439529453427565_562257601125925561_n.jpg?_nc_ht=scontent-frx5-1.cdninstagram.com'
                '&_nc_cat=101&oh=4ae866a0b55b092ef9868e74f19b5623&oe=5E8BA669'
        ))
        expected_images.add(Image(
            source='https://www.instagram.com/p/B5F4e14AO9k/',
            uri='https://scontent-frx5-1.cdninstagram.com/v/t51.2885-15/e35/73480811_2530202687056717'
                '_5816769842868095145_n.jpg?_nc_ht=scontent-frx5-1.cdninstagram.com'
                '&_nc_cat=101&oh=2515a15280e2d1602f672593c9f0da29&oe=5E5116A9'
        ))
        expected_images.add(Image(
            source='https://www.instagram.com/p/B5F27k1BxQG/',
            uri='https://scontent-frx5-1.cdninstagram.com/v/t51.2885-15/e35/71318465_578947536241820_'
                '8770680984207687503_n.jpg?_nc_ht=scontent-frx5-1.cdninstagram.com&_nc_cat=108&'
                'oh=8612bc5a2e5ad265d885dbeb5861741c&oe=5E81D4BE'
        ))
        expected_images.add(Image(
            source='https://www.instagram.com/p/B5Fz_oqDbrY/',
            uri='https://scontent-frx5-1.cdninstagram.com/v/t51.2885-15/e35/73171316_154443005781309_'
                '5952715908442524817_n.jpg?_nc_ht=scontent-frx5-1.cdninstagram.com&_nc_cat=109'
                '&oh=61633aa07eb05b103b799f65a21bd60a&oe=5E6D0DFF'
        ))
        expected_images.add(Image(
            source='https://www.instagram.com/p/B5Fzi5oHY-O/',
            uri='https://scontent-frx5-1.cdninstagram.com/v/t51.2885-15/e35/74600030_100916318024628_'
                '8288469442916469948_n.jpg?_nc_ht=scontent-frx5-1.cdninstagram.com&_nc_cat=100'
                '&oh=a8d237a032264976286879a581c0d904&oe=5E4FD98D'
        ))
        expected_images.add(Image(
            source='https://www.instagram.com/p/B5Fu-2CgcoG/',
            uri='https://scontent-frx5-1.cdninstagram.com/v/t51.2885-15/e35/p1080x1080/77151874_'
                '149403059703505_1673985012081769188_n.jpg?_nc_ht=scontent-frx5-1.cdninstagram.com'
                '&_nc_cat=101&oh=496e626094a5dc32b60eea9f5785a814&oe=5E4D8D42'
        ))
        expected_images.add(Image(
            source='https://www.instagram.com/p/B5FssMBFyUL/',
            uri='https://scontent-frx5-1.cdninstagram.com/v/t51.2885-15/e35/73381083_998377233839407_'
                '3873966681762746452_n.jpg?_nc_ht=scontent-frx5-1.cdninstagram.com&_nc_cat=107'
                '&oh=651b7a89927a465d758679c599c9aa18&oe=5E687FE2'
        ))
        expected_images.add(Image(
            source='https://www.instagram.com/p/B5FpfZ2B4t1/',
            uri='https://scontent-frx5-1.cdninstagram.com/v/t51.2885-15/e35/73300147_1391854841147986_'
                '2913574931708576695_n.jpg?_nc_ht=scontent-frx5-1.cdninstagram.com&_nc_cat=110'
                '&oh=18800daa9416377ed6b870acc0502e2a&oe=5E6E8495'
        ))
        # act
        images = self.crawler._crawl()
        # assert
        self.assertEqual(self.crawler._cursor, expected_cursor)
        self.assertSetEqual(images, expected_images)
        for expected_image in expected_images:
            for image in images:
                if image == expected_image:
                    # sources are irrelevant for equality, need to be checked manually
                    self.assertEqual(image.source, expected_image.source)


class InstagramHashtagDescriptionTest(unittest.TestCase):
    def test_description_config(self) -> None:
        # act
        description = InstagramHashtag.info()
        # assert
        assert isinstance(description.config, dict)
        self.assertIn('tag_name', description.config)


class InstagramHashtagLoaderTest(ImageCrawlerLoaderTest):

    @property
    def ic_name(self) -> str:
        return 'InstagramHashtag'

    @property
    def ic_class(self) -> Type[BaseImageCrawler]:
        return InstagramHashtag

    def test_loader(self) -> None:
        self.check()


class InstagramProfileTest(unittest.TestCase):

    _PROFILE_ID = '787132'
    _QUERY_HASH = '2c5d4d8b70cad329c4a6ebe3abb6eedd'

    def setUp(self) -> None:
        InstagramProfile._query_hash = None
        self.crawler = InstagramProfile(user_name='natgeo')
        self.crawler._remote_fetcher = _FILE_FETCHER
        self.crawler._amount = 5
        self.crawler._get_queryhashfinder = self._get_queryhashfinder  # type: ignore

    def tearDown(self) -> None:
        del self.crawler
        InstagramProfile._query_hash = None

    def _get_queryhashfinder(self) -> InstagramQueryHashFinder:
        finder = InstagramQueryHashFinder('profile')
        finder.find_hashes = lambda: _QUERYHASHES_EXPECTED_PROFILE  # type: ignore
        return finder

    def _get_profile_id(self) -> str:
        return self.__class__._PROFILE_ID

    def test__get_profile_id(self) -> None:
        # act
        profile_id = self.crawler._get_profile_id()
        # assert
        self.assertEqual(profile_id, self.__class__._PROFILE_ID)

    def test__get_query_hash(self) -> None:
        # arrange
        self.crawler._get_profile_id = self._get_profile_id  # type: ignore
        # act
        queryhash = self.crawler._get_query_hash()
        # assert
        self.assertEqual(queryhash, self.__class__._QUERY_HASH)

    def test__crawl(self) -> None:
        # arrange
        self.crawler._get_query_hash = lambda: self.__class__._QUERY_HASH  # type: ignore
        self.crawler._get_profile_id = self._get_profile_id  # type: ignore
        expected_images = ImageCollection()
        expected_cursor = 'QVFBbjdTc0dOQ2JQTW1vb1JzMTQxeGpkMEFnTzhYWmh5dFRfMXRWT1VwX28' \
                          'wMUxkSExpZ2s5SVZfWmM5VWtjYUJrTS0wTW5Va2JqSEpTSUpPcENnN1g1OQ=='
        expected_images.add(Image(
            uri='https://scontent-frt3-1.cdninstagram.com/v/t51.2885-15/e35/75467914_150782276185354_'
                '1270489924400076442_n.jpg?_nc_ht=scontent-frt3-1.cdninstagram.com&'
                '_nc_cat=1&oh=f293b2a234c82263dfd37b3785e19625&oe=5E812486',
            source='https://www.instagram.com/p/B5GWlkfjgWZ/'
        ))
        expected_images.add(Image(
            uri='https://scontent-frt3-1.cdninstagram.com/v/t51.2885-15/e35/75467914_150782276185354_'
                '1270489924400076442_n.jpg?_nc_ht=scontent-frt3-1.cdninstagram.com&_nc_cat=1&'
                'oh=f293b2a234c82263dfd37b3785e19625&oe=5E812486',
            source='https://www.instagram.com/p/B5GWlkfjgWZ/'
        ))
        expected_images.add(Image(
            uri='https://scontent-frt3-1.cdninstagram.com/v/t51.2885-15/e35/75349300_522545705143892_'
                '7892885809773901918_n.jpg?_nc_ht=scontent-frt3-1.cdninstagram.com&_nc_cat=1&'
                'oh=1d2f00065294117027b17a585b7d05ab&oe=5E4E791D',
            source='https://www.instagram.com/p/B5GWlkfjgWZ/'
        ))
        expected_images.add(Image(
            uri='https://scontent-frt3-1.cdninstagram.com/v/t51.2885-15/e35/73512651_760344041101107_'
                '8305449940174698639_n.jpg?_nc_ht=scontent-frt3-1.cdninstagram.com&'
                '_nc_cat=1&oh=cc60db8291b59c26af119139ec130edf&oe=5E6B47D3',
            source='https://www.instagram.com/p/B5GWlkfjgWZ/'
        ))
        expected_images.add(Image(
            uri='https://scontent-frt3-1.cdninstagram.com/v/t51.2885-15/e35/s1080x1080/75349296'
                '_565205657598924_1414110584006258216_n.jpg?_nc_ht=scontent-frt3-1.cdninstagram.com&'
                '_nc_cat=1&oh=5cd52e00ce5b9571013820950d32a9db&oe=5E4E8A5F',
            source='https://www.instagram.com/p/B5GCEIsj-Wl/'
        ))
        expected_images.add(Image(
            uri='https://scontent-frt3-1.cdninstagram.com/v/t51.2885-15/e35/s1080x1080/75349296_'
                '565205657598924_1414110584006258216_n.jpg?_nc_ht=scontent-frt3-1.cdninstagram.com&'
                '_nc_cat=1&oh=5cd52e00ce5b9571013820950d32a9db&oe=5E4E8A5F',
            source='https://www.instagram.com/p/B5GCEIsj-Wl/'
        ))
        expected_images.add(Image(
            uri='https://scontent-frt3-1.cdninstagram.com/v/t51.2885-15/e35/s1080x1080/70023824_'
                '3272578002812323_8917281619820840144_n.jpg?_nc_ht=scontent-frt3-1.cdninstagram.com&'
                '_nc_cat=1&oh=46773eb8dbebc8db8a5acf07e0d9ee94&oe=5E539F29',
            source='https://www.instagram.com/p/B5GCEIsj-Wl/'
        ))
        expected_images.add(Image(
            uri='https://scontent-frt3-1.cdninstagram.com/v/t51.2885-15/e35/s1080x1080/73480790_830405794055849_'
                '4155404398603377777_n.jpg?_nc_ht=scontent-frt3-1.cdninstagram.com&_nc_cat=1&'
                'oh=5ab282415040be2108c1e0e5fadf8a2a&oe=5E50079A',
            source='https://www.instagram.com/p/B5GCEIsj-Wl/'
        ))
        expected_images.add(Image(
            uri='https://scontent-frt3-1.cdninstagram.com/v/t51.2885-15/e35/s1080x1080/74350656_146186010045632_'
                '5113331273863195582_n.jpg?_nc_ht=scontent-frt3-1.cdninstagram.com&_nc_cat=1&'
                'oh=67f218bed182ad8bc7ff9118b2d88139&oe=5E5664A6',
            source='https://www.instagram.com/p/B5GCEIsj-Wl/'
        ))
        expected_images.add(Image(
            uri='https://scontent-frt3-1.cdninstagram.com/v/t51.2885-15/e35/s1080x1080/75375719_164423041289362_'
                '4914559208372349272_n.jpg?_nc_ht=scontent-frt3-1.cdninstagram.com&_nc_cat=1&'
                'oh=b401628c44cf7155f6e7963872782306&oe=5E52153E',
            source='https://www.instagram.com/p/B5GCEIsj-Wl/'
        ))
        expected_images.add(Image(
            uri='https://scontent-frt3-1.cdninstagram.com/v/t51.2885-15/e35/s1080x1080/75538167_2161689137265601_'
                '3450507258498841854_n.jpg?_nc_ht=scontent-frt3-1.cdninstagram.com&'
                '_nc_cat=1&oh=484ebc5923e3755f4b88a9cde4d01e37&oe=5E6CE284',
            source='https://www.instagram.com/p/B5GCEIsj-Wl/'
        ))
        expected_images.add(Image(
            uri='https://scontent-frt3-1.cdninstagram.com/v/t51.2885-15/fr/e15/s1080x1080/73401893_805717106534521_'
                '4743540237997542732_n.jpg?_nc_ht=scontent-frt3-1.cdninstagram.com&_nc_cat=1&'
                'oh=a036d5fbe716cd7b11d313200e5ba73d&oe=5E538C0A',
            source='https://www.instagram.com/p/B5FY2u9j8nG/'
        ))
        expected_images.add(Image(
            uri='https://scontent-frt3-1.cdninstagram.com/v/t51.2885-15/fr/e15/s1080x1080/74600036_170329604158818_'
                '2739654930228765968_n.jpg?_nc_ht=scontent-frt3-1.cdninstagram.com&_nc_cat=1&'
                'oh=84732a7538ab59d3bd1458d400e181fa&oe=5E86FC97',
            source='https://www.instagram.com/p/B5FEWeHAT-M/'
        ))

        # act
        images = self.crawler._crawl()
        # assert
        self.assertEqual(self.crawler._cursor, expected_cursor)
        self.assertSetEqual(images, expected_images)
        for expected_image in expected_images:
            for image in images:
                if image == expected_image:
                    # sources are irrelevant for equality, need to be checked manually
                    self.assertEqual(image.source, expected_image.source)

    def test_reset(self) -> None:
        # arrange
        self.crawler._cursor = 'foo'
        # act
        self.crawler._reset()
        # assert
        self.assertIsNone(self.crawler._cursor)


class InstagramProfileDescriptionTest(unittest.TestCase):

    def test_description_config(self) -> None:
        # act
        description = InstagramProfile.info()
        # assert
        assert isinstance(description.config, dict)
        self.assertIn('user_name', description.config)


class InstagramProfileLoaderTest(ImageCrawlerLoaderTest):

    @property
    def ic_name(self) -> str:
        return 'InstagramProfile'

    @property
    def ic_class(self) -> Type[BaseImageCrawler]:
        return InstagramProfile

    def test_loader(self) -> None:
        self.check()



import unittest


class CommandsRunTest(unittest.TestCase):

    def test_(self) -> None:
        self.skipTest('TODO')


class CommandsConfigTest(unittest.TestCase):

    def test_(self) -> None:
        self.skipTest('TODO')


class CommandsConfigDumpFileTest(unittest.TestCase):

    def test_(self) -> None:
        self.skipTest('TODO')


class CommandsConfigCheckFileTest(unittest.TestCase):

    def test_(self) -> None:
        self.skipTest('TODO')


class CommandsInfoTest(unittest.TestCase):

    def test_(self) -> None:
        self.skipTest('TODO')


class CommandsInfoVersionTest(unittest.TestCase):

    def test_(self) -> None:
        self.skipTest('TODO')


class CommandsInfoImagecrawlerListTest(unittest.TestCase):

    def test_(self) -> None:
        self.skipTest('TODO')


class CommandsInfoImagecrawlerDescTest(unittest.TestCase):

    def test_(self) -> None:
        self.skipTest('TODO')



from glob import glob
from os.path import basename, dirname, join as path_join, realpath

import pytest  # type: ignore
from ddt import ddt, idata, unpack  # type: ignore

from nichtparasoup.testing.config import ConfigFileTest


@ddt
class ShippedExampleConfigFileTest(ConfigFileTest):
    __EXAMPLE_DIR = path_join(dirname(__file__), '..', '..', '..', 'examples', 'config')
    __EXAMPLE_FILES = glob(path_join(__EXAMPLE_DIR, '*.*'), recursive=False)

    def test_has_examples(self) -> None:
        self.assertGreater(len(self.__class__.__EXAMPLE_FILES), 0)

    @pytest.mark.online  # type: ignore  # probing goes over the internet
    @pytest.mark.slow  # type: ignore  # probing might be slow
    @idata((basename(file),) for file in __EXAMPLE_FILES)  # type: ignore
    @unpack  # type: ignore
    def test_example(self, filename: str) -> None:
        file = realpath(path_join(self.__class__.__EXAMPLE_DIR, filename))
        self.validate(file)
        self.probe(file)



from typing import Any, Dict

from nichtparasoup.imagecrawler import BaseImageCrawler, Image, ImageCollection, ImageCrawlerConfig, ImageCrawlerInfo

__all__ = ["Picsum"]


class Picsum(BaseImageCrawler):
    _bunch = 10

    @classmethod
    def info(cls) -> ImageCrawlerInfo:
        return ImageCrawlerInfo(
            description='Find images from https://picsum.photos',
            config=dict(
                width='how many pixels of the image to find should be wide',
                height='how many pixels of the image to find should be high',
            ),
            icon_url='https://picsum.photos/assets/images/favicon/favicon-32x32.png',
        )

    @classmethod
    def check_config(cls, config: Dict[Any, Any]) -> ImageCrawlerConfig:
        width = config["width"]
        height = config["height"]
        if type(width) is not int:
            raise TypeError("width {!r} is not int".format(width))
        if type(height) is not int:
            raise TypeError("height {!r} is not int".format(height))
        if width <= 0:
            raise ValueError("width {!r} <= 0".format(width))
        if height <= 0:
            raise ValueError("height {!r} <= 0".format(height))
        return ImageCrawlerConfig(
            width=width,
            height=height,
        )

    @staticmethod
    def _get_image_uri(width: int, height: int) -> str:
        return "https://picsum.photos/{}/{}".format(width, height)

    def _reset(self) -> None:  # pragma: no cover
        pass

    def _crawl(self) -> ImageCollection:
        images = ImageCollection()
        config = self.get_config()
        for _ in range(0, self._bunch):
            uri = self._get_image_uri(**config)
            images.add(Image(
                uri=uri,
                source=uri,
                is_generic=True,
            ))
        return images



__all__ = ["Reddit"]

from json import loads as json_loads
from typing import Any, Dict, Optional
from urllib.parse import quote_plus as url_quote, urljoin

from nichtparasoup.imagecrawler import (
    BaseImageCrawler, Image, ImageCollection, ImageCrawlerConfig, ImageCrawlerInfo, ImageRecognizer, RemoteFetcher,
)


class Reddit(BaseImageCrawler):

    def __init__(self, **config: Any) -> None:  # pragma: no cover
        super().__init__(**config)
        self._uri_base = 'https://www.reddit.com/r/{}.json?after='.format(
            url_quote(self._config['subreddit']))
        self._after = None  # type: Optional[str]
        self._remote_fetcher = RemoteFetcher()
        self._image_recognizer = ImageRecognizer()

    @classmethod
    def info(cls) -> ImageCrawlerInfo:
        return ImageCrawlerInfo(
            description='A Crawler for an arbitrary SubReddit of https://www.reddit.com',
            config=dict(
                subreddit='the SubReddit to crawl',
            ),
            icon_url='https://www.redditstatic.com/desktop2x/img/favicon/apple-icon-120x120.png',
        )

    @classmethod
    def check_config(cls, config: Dict[Any, Any]) -> ImageCrawlerConfig:
        subreddit = config["subreddit"]
        if type(subreddit) is not str:
            raise TypeError("subreddit {!r} is not str".format(subreddit))
        if 0 == len(subreddit):
            raise ValueError("subreddit {!r} is empty".format(subreddit))
        return ImageCrawlerConfig(
            subreddit=subreddit,
        )

    def _reset(self) -> None:
        self._after = None

    def _crawl(self) -> ImageCollection:
        images = ImageCollection()
        listing_string, uri = self._remote_fetcher.get_string(self._get_uri(self._after))
        listing = json_loads(listing_string)
        del listing_string  # free up some ram
        for child in listing['data']['children']:
            image = self._get_image(child['data'])
            if image:
                images.add(Image(
                    uri=image,
                    source=urljoin(uri, child['data']['permalink']),
                ))
        # don't care if `after` is `None` after the crawl ... why not restarting at front when the end is reached?!
        self._after = listing['data']['after']
        return images

    def _get_uri(self, after: Optional[str]) -> str:
        return self._uri_base + (url_quote(after) if after else '')

    def _get_image(self, data: Dict[str, Any]) -> Optional[str]:
        uri = data.get('url')  # type: Optional[str]
        return uri if uri and self._image_recognizer.path_is_image(uri) else None



import unittest


class VersionTest(unittest.TestCase):

    def test_version_known(self) -> None:
        # arrange
        from nichtparasoup import VERSION
        # assert
        self.assertFalse(
            VERSION.startswith('UNKNOWN'),
            'unknown version: {!r}'.format(VERSION)
        )



__all__ = ["get_config", "get_defaults", "dump_defaults", "get_imagecrawler", "parse_yaml_file",
           "ImageCrawlerSetupError"]

from os.path import dirname, join as path_join, realpath
from typing import Any, Dict, Optional

from nichtparasoup.core.imagecrawler import BaseImageCrawler

_SCHEMA_FILE = realpath(path_join(dirname(__file__), "schema.yaml"))
_schema = None  # type: Optional[Any]

_DEFAULTS_FILE = realpath(path_join(dirname(__file__), "defaults.yaml"))
_defaults = None  # type: Optional[Dict[str, Any]]


class ImageCrawlerSetupError(Exception):

    def __init__(self, ic_name: str, ic_class: type, ic_config: Dict[Any, Any]) -> None:  # pragma: no cover
        self._name = ic_name
        self._class = ic_class
        self._config = ic_config

    def __str__(self) -> str:  # pragma: no cover
        return 'Failed setup crawler {!r} of type {!r} with config {!r}'.format(self._name, self._class, self._config)


def get_imagecrawler(config_crawler: Dict[str, Any]) -> BaseImageCrawler:
    from nichtparasoup.imagecrawler import get_imagecrawlers
    imagecrawler_name = config_crawler['name']
    imagecrawler_class = get_imagecrawlers().get_class(imagecrawler_name)
    if not imagecrawler_class:
        raise ValueError('unknown crawler name {!r}'.format(imagecrawler_name))
    imagecrawler_config = config_crawler['config']
    try:
        return imagecrawler_class(**imagecrawler_config)
    except Exception as e:
        raise ImageCrawlerSetupError(imagecrawler_name, imagecrawler_class, imagecrawler_config) from e


def parse_yaml_file(file_path: str) -> Dict[str, Any]:
    import yamale  # type: ignore
    global _schema
    if not _schema:
        _schema = yamale.make_schema(_SCHEMA_FILE, parser='ruamel')
    _data = yamale.make_data(file_path, parser='ruamel')
    config = yamale.validate(_schema, _data, strict=True)[0][0]  # type: Dict[str, Any]
    config.setdefault('logging', dict())
    config['logging'].setdefault('level', 'INFO')
    for config_crawler in config['crawlers']:
        config_crawler.setdefault("weight", 1)
        config_crawler.setdefault('config', dict())
    return config


def dump_defaults(file_path: str) -> None:
    from shutil import copyfile
    copyfile(_DEFAULTS_FILE, file_path)


def get_defaults() -> Dict[str, Any]:
    global _defaults
    if not _defaults:
        _defaults = parse_yaml_file(_DEFAULTS_FILE)
    from copy import deepcopy
    return deepcopy(_defaults)


def get_config(config_file: Optional[str] = None) -> Dict[str, Any]:
    if not config_file:
        return get_defaults()
    try:
        return parse_yaml_file(config_file)
    except Exception as e:
        raise ValueError('invalid config file {!r}'.format(config_file)) from e



import unittest

from nichtparasoup.core import NPCore
from nichtparasoup.core.image import Image

from .mockable_imagecrawler import MockableImageCrawler


class NPCoreTest(unittest.TestCase):

    def test__is_image_not_in_blacklist(self) -> None:
        # arrange
        image1 = Image(uri="test1", source='test')
        image2 = Image(uri="test2", source='test')
        core = NPCore()
        # act
        core.blacklist.add(image1.uri)
        # assert
        self.assertTrue(core._is_image_not_in_blacklist(image2), 'expected image in blacklist')
        self.assertFalse(core._is_image_not_in_blacklist(image1), 'expected image not in blacklist')

    def test__add_image_to_blacklist(self) -> None:
        # arrange
        image1 = Image(uri="test1", source='test')
        image2 = Image(uri="test2", source='test')
        core = NPCore()
        # act
        core._add_image_to_blacklist(image1)
        # assert
        self.assertIn(image1.uri, core.blacklist)
        self.assertNotIn(image2.uri, core.blacklist)

    def test_add_and_has_imagecrawler(self) -> None:
        # arrange
        core = NPCore()
        imagecrawler1 = MockableImageCrawler(foo='bar')
        imagecrawler2 = MockableImageCrawler(bar='bazz')
        imagecrawler3 = MockableImageCrawler(bar='bazz')
        # act
        core.add_imagecrawler(imagecrawler1, 1)
        core.add_imagecrawler(imagecrawler2, 1)
        # assert
        self.assertEqual(2, len(core.crawlers))
        self.assertListEqual(
            [imagecrawler1, imagecrawler2],
            [crawler.imagecrawler for crawler in core.crawlers])
        self.assertTrue(core.has_imagecrawler(imagecrawler1))
        self.assertTrue(core.has_imagecrawler(imagecrawler3))



__all__ = ["Server", "ServerStatistics", "ServerStatus", "ServerRefiller"]

from abc import ABC
from copy import copy
from sys import getsizeof
from threading import Lock, Thread
from time import time
from typing import Any, Dict, Optional, Type, Union
from weakref import ref as weak_ref

from nichtparasoup import VERSION
from nichtparasoup._internals import _log
from nichtparasoup.core import Crawler, NPCore


class Server(object):
    """
    this class is intended to be thread save.
    this class intended to be a stable interface.
    its public methods return base types only.
    """

    def __init__(self, core: NPCore, crawler_upkeep: int = 30,
                 reset_timeout: int = 60 * 60) -> None:  # pragma: no cover
        self.core = core
        self.keep = crawler_upkeep
        self.reset_timeout = reset_timeout
        self._stats = ServerStatistics()
        self._refiller = None  # type: Optional[ServerRefiller]
        self._trigger_reset = False
        self._locks = _ServerLocks()
        self.__running = False

    def get_image(self) -> Optional[Dict[str, Any]]:
        crawler = self.core.crawlers.get_random()
        if not crawler:
            return None
        image = copy(crawler.pop_random_image())
        if not image:
            return None
        with self._locks.stats_get_image:
            self._stats.count_images_served += 1
        return dict(
            uri=image.uri,
            is_generic=image.is_generic,
            source=image.source,
            more=image.more,
            crawler=dict(
                id=id(crawler),
                type=type_module_name_str(type(crawler.imagecrawler)),
            ),
        )

    @staticmethod
    def _log_refill_crawler(crawler: Crawler, refilled: int) -> None:
        # must be compatible to nichtparasoup.core._OnFill
        if refilled > 0:
            _log('info', "refilled by {} via {!r}".format(refilled, crawler.imagecrawler))

    def refill(self) -> Dict[str, bool]:
        with self._locks.refill:
            self.core.fill_up_to(self.keep, self._log_refill_crawler)
            return dict(refilled=True)

    def _reset(self) -> None:
        with self._locks.reset:
            self._stats.cum_blacklist_on_flush += self.core.reset()
            self._stats.count_reset += 1
            self._stats.time_last_reset = int(time())

    def request_reset(self) -> Dict[str, Any]:
        if not self.is_alive():
            request_valid = True
            timeout = 0
        else:
            now = int(time())
            time_started = self._stats.time_started or now
            timeout_base = self.reset_timeout
            time_last_reset = self._stats.time_last_reset
            reset_after = timeout_base + (time_last_reset or time_started)
            request_valid = now > reset_after
            timeout = timeout_base if request_valid else (reset_after - now)
        if request_valid:
            self._reset()
        return dict(
            requested=request_valid,
            timeout=timeout,
        )

    def start(self) -> None:
        with self._locks.run:
            if self.__running:
                raise RuntimeError('already running')
            _log('info', " * starting {}".format(type(self).__name__))
            _log('info', ' * fill all crawlers up to {}'.format(self.keep))
            self.refill()  # initial fill
            if not self._refiller:
                self._refiller = ServerRefiller(self, 1)
                self._refiller.start()  # start threaded periodical refill
            self._stats.time_started = int(time())
            self.__running = True

    def is_alive(self) -> bool:
        return self.__running

    def stop(self) -> None:
        with self._locks.run:
            if not self.__running:
                raise RuntimeError('not running')
            _log('info', "\r\n * stopping {}".format(type(self).__name__))
            if self._refiller:
                self._refiller.stop()
                self._refiller = None
            self.__running = False


class ServerStatus(ABC):
    """
    this class intended to be a stable interface.
    all public methods are like this: Callable[[Server], Union[List[SomeBaseType], Dict[str, SomeBaseType]]]
    all public methods must be associated with stat(u)s!
    """

    @staticmethod
    def server(server: Server) -> Dict[str, Any]:
        stats = copy(server._stats)
        now = int(time())
        uptime = (now - stats.time_started) if server.is_alive() and stats.time_started else 0
        return dict(
            version=VERSION,
            uptime=uptime,
            reset=dict(
                count=stats.count_reset,
                since=(now - stats.time_last_reset) if stats.time_last_reset else uptime,
            ),
            images=dict(
                served=stats.count_images_served,
                crawled=stats.cum_blacklist_on_flush + len(server.core.blacklist),
            ),
        )

    @staticmethod
    def blacklist(server: Server) -> Dict[str, Any]:
        blacklist = server.core.blacklist.copy()
        return dict(
            len=len(blacklist),
            size=getsizeof(blacklist),
        )

    @staticmethod
    def crawlers(server: Server) -> Dict[int, Dict[str, Any]]:
        status = dict()
        for crawler in server.core.crawlers.copy():
            crawler_id = id(crawler)
            crawler = copy(crawler)
            images = crawler.images.copy()
            status[crawler_id] = dict(
                weight=crawler.weight,
                type=type_module_name_str(type(crawler.imagecrawler)),
                config=crawler.imagecrawler.get_config(),  # just a dict
                images=dict(
                    len=len(images),
                    size=getsizeof(images),
                ),
            )
        return status


def type_module_name_str(t: Type[Any]) -> str:
    return '{}:{}'.format(t.__module__, t.__name__)


class ServerRefiller(Thread):
    def __init__(self, server: Server, sleep: Union[int, float]) -> None:  # pragma: no cover
        from threading import Event
        super().__init__(daemon=True)
        self._server_wr = weak_ref(server)
        self._sleep = sleep
        self._stop_event = Event()
        self._run_lock = Lock()

    def run(self) -> None:
        from time import sleep
        while not self._stop_event.is_set():
            server = self._server_wr()  # type: Optional[Server]
            if server:
                server.refill()
            else:
                _log('info', " * server gone. stopping {}".format(type(self).__name__))
                self._stop_event.set()
            if not self._stop_event.is_set():
                sleep(self._sleep)

    def start(self) -> None:
        self._run_lock.acquire()
        try:
            if self.is_alive():
                raise RuntimeError('already running')
            _log('info', " * starting {}".format(type(self).__name__))
            self._stop_event.clear()
            super().start()
        finally:
            self._run_lock.release()

    def stop(self) -> None:
        self._run_lock.acquire()
        try:
            if not self.is_alive():
                raise RuntimeError('not running')
            _log('info', " * stopping {}".format(type(self).__name__))
            self._stop_event.set()
        finally:
            self._run_lock.release()


class ServerStatistics(object):
    def __init__(self) -> None:  # pragma: no cover
        self.time_started = None  # type: Optional[int]
        self.count_images_served = 0  # type: int
        self.count_reset = 0  # type: int
        self.time_last_reset = None  # type: Optional[int]
        self.cum_blacklist_on_flush = 0  # type: int


class _ServerLocks(object):
    def __init__(self) -> None:  # pragma: no cover
        self.stats_get_image = Lock()
        self.reset = Lock()
        self.refill = Lock()
        self.run = Lock()



import unittest
from os import path

from nichtparasoup.testing.config import ConfigFileTest


class ConfigFileValidateTest(unittest.TestCase):

    _TESTDATA_DIR = path.join(path.dirname(__file__), 'testdata_configfiletest')

    def test_propagate_errors(self) -> None:
        # nothing to test, yet.
        # calls already covered functions only
        pass

    def test_detect_duplicates(self) -> None:
        import re
        # arrange
        tester = ConfigFileTest()
        test_file = path.join(self._TESTDATA_DIR, 'duplicates.yaml')
        # act & assert
        with self.assertRaisesRegex(Exception, re.compile(r'duplicate ImageCrawler', re.IGNORECASE)):
            tester.validate(test_file)


class ConfigFileProbeTest(unittest.TestCase):

    def test_propagate_errors(self) -> None:
        # nothing to test, yet.
        # calls already covered functions only
        pass



import unittest

from nichtparasoup.core import NPCore
from nichtparasoup.core.server import Server
from nichtparasoup.webserver import WebServer


class WebserverTest(unittest.TestCase):

    def setUp(self) -> None:
        self.webserver = WebServer(Server(NPCore()), "", 0)

    def tearDown(self) -> None:
        del self.webserver

    def test_webserver(self) -> None:
        self.skipTest('TODO: write the test')  # TODO: write tests
        # maybe this helps? http://werkzeug.palletsprojects.com/en/0.16.x/test/



__all__ = [
    "get_imagecrawlers",
    # for convenience, all classes that are needed to implement an ImageCrawler are exported, here
    "BaseImageCrawler", "ImageCrawlerConfig", "ImageCrawlerInfo",
    "Image", "ImageCollection",
    "RemoteFetcher", "ImageRecognizer",
]

from copy import copy
from typing import Any, Dict, Iterable, List, Optional, Tuple, Type

from pkg_resources import EntryPoint, iter_entry_points

from nichtparasoup._internals import _log
from nichtparasoup.core.image import Image, ImageCollection
from nichtparasoup.core.imagecrawler import (
    BaseImageCrawler, ImageCrawlerConfig, ImageCrawlerInfo, ImageRecognizer, RemoteFetcher,
)

_ImagecrawlerName = str
_ImagecrawlerClass = Type[BaseImageCrawler]
_Imagecrawler = Tuple[_ImagecrawlerName, _ImagecrawlerClass]


class KnownImageCrawlers(object):

    @staticmethod
    def _builtins() -> Dict[_ImagecrawlerName, _ImagecrawlerClass]:
        from .echo import Echo
        from .picsum import Picsum
        from .reddit import Reddit
        from .instagram import InstagramHashtag, InstagramProfile
        from .pr0gramm import Pr0gramm
        return dict(
            Echo=Echo,
            Picsum=Picsum,
            Reddit=Reddit,
            InstagramProfile=InstagramProfile,
            InstagramHashtag=InstagramHashtag,
            Pr0gramm=Pr0gramm,
        )

    def __init__(self, entries: Iterable[EntryPoint]) -> None:  # pragma: no cover
        self._list = [(n, c) for n, c in self._builtins().items()]  # type: List[_Imagecrawler]
        for entry in entries:
            try:
                self._add(entry)
                _log('debug', 'Entry point added: {} from {!r}'.format(entry, entry.dist))
            except Exception as e:
                _log('debug', 'Entry point skipped: {} from {!r}\r\n\t{}'.format(entry, entry.dist, e), exc_info=True)

    def _add(self, entry: EntryPoint) -> None:
        self._test_duplicate_name(entry.name)
        loaded = self._load(entry)
        self._test(loaded)
        self._test_duplicate_class(loaded)
        # if everything went well .. add
        self._list.append((entry.name, loaded))

    def names(self) -> Iterable[_ImagecrawlerName]:
        return (ic_name for ic_name, _ in self._list)

    def classes(self) -> Iterable[_ImagecrawlerClass]:
        return (ic_class for _, ic_class in self._list)

    def items(self) -> Iterable[_Imagecrawler]:
        return (copy(imacrawler) for imacrawler in self._list)

    def get_name(self, imagecrawler_class: _ImagecrawlerClass) -> Optional[_ImagecrawlerName]:
        for ic_name, ic_class in self._list:
            if ic_class == imagecrawler_class:
                return ic_name
        return None

    def get_class(self, imagecrawler_name: _ImagecrawlerName) -> Optional[_ImagecrawlerClass]:
        for ic_name, ic_class in self._list:
            if ic_name == imagecrawler_name:
                return ic_class
        return None

    @staticmethod
    def _load(entry: EntryPoint) -> Any:
        try:
            return entry.load()
        except Exception as e:
            raise ImportError('Error on loading entry {}'.format(entry)) from e

    @classmethod
    def _test(cls, something: Any) -> None:  # pragma: no cover
        cls._test_inheritance(something)
        cls._test_abstract(something)

    @staticmethod
    def _test_inheritance(something: Any) -> None:
        if not issubclass(something, BaseImageCrawler):
            raise TypeError('{!r} is not a {!r}'.format(something, BaseImageCrawler))

    @staticmethod
    def _test_abstract(some_type: Type[object]) -> None:
        if hasattr(some_type, '__abstractmethods__') and some_type.__abstractmethods__:  # type: ignore
            raise TypeError('{!r} is abstract'.format(some_type))

    def _test_duplicate_name(self, imagecrawler_name: str) -> None:
        for ic_name, _ in self._list:
            if ic_name == imagecrawler_name:
                raise KeyError('Duplicate ImageCrawler {!r}'.format(imagecrawler_name))

    def _test_duplicate_class(self, imagecrawler_class: _ImagecrawlerClass) -> None:
        for _, ic_class in self._list:
            if ic_class == imagecrawler_class:
                raise TypeError('Duplicate ImageCrawler {!r}'.format(imagecrawler_class))


__ENTRY_POINT_NAME = 'nichtparasoup_imagecrawler'

__imagecrawlers = None  # type: Optional[KnownImageCrawlers]


def get_imagecrawlers() -> KnownImageCrawlers:  # pragma: no cover
    global __imagecrawlers
    if __imagecrawlers is None:
        __imagecrawlers = KnownImageCrawlers(iter_entry_points(__ENTRY_POINT_NAME))
    return __imagecrawlers


def clear_imagecrawlers() -> None:  # pragma: no cover
    global __imagecrawlers
    __imagecrawlers = None



import unittest
from os.path import dirname, join as path_join
from typing import Type

from nichtparasoup.imagecrawler import BaseImageCrawler, Image, ImageCollection
from nichtparasoup.imagecrawler.reddit import Reddit
from nichtparasoup.testing.imagecrawler import FileFetcher, ImageCrawlerLoaderTest

_reddit_right_config = dict(subreddit='aww')


class RedditConfigCorrect(unittest.TestCase):

    def test__check_config_right_value(self) -> None:
        # arrange
        config_in = _reddit_right_config.copy()
        # act
        config_out = Reddit.check_config(config_in)
        # assert
        self.assertDictEqual(config_in, config_out)


class RedditConfigSubredditTest(unittest.TestCase):

    def setUp(self) -> None:
        self._reddit_right_config_wo_subreddit = _reddit_right_config.copy()
        del self._reddit_right_config_wo_subreddit["subreddit"]

    def tearDown(self) -> None:
        del self._reddit_right_config_wo_subreddit

    def test_check_config_missing_value(self) -> None:
        # assert
        with self.assertRaises(KeyError):
            Reddit.check_config(self._reddit_right_config_wo_subreddit)

    def test_check_config_wrong_type(self) -> None:
        wrong_types = [None, True, 1, 0.1, [], (), {}, self]  # type: ignore
        for wrong_type in wrong_types:
            # arrange
            config_in = self._reddit_right_config_wo_subreddit
            config_in["subreddit"] = wrong_type  # type: ignore
            # assert
            with self.assertRaises(TypeError, msg=repr(config_in)):
                Reddit.check_config(config_in)

    def test_check_config_wrong_value(self) -> None:
        wrong_values = [""]
        for wrong_value in wrong_values:
            # arrange
            config_in = self._reddit_right_config_wo_subreddit
            config_in["subreddit"] = wrong_value
            # assert
            with self.assertRaises(ValueError, msg=repr(config_in)):
                Reddit.check_config(config_in)

    def test_check_config_not_existing(self) -> None:
        self.skipTest('TODO do we need this tested or even coded?')


class RedditBuildUriTest(unittest.TestCase):

    def test__build_uri_at_front(self) -> None:
        # arrange
        crawler = Reddit(subreddit='foo')
        # act
        uri = crawler._get_uri(None)
        # assert
        self.assertEqual(uri, 'https://www.reddit.com/r/foo.json?after=')

    def test__build_uri_at_front__escape(self) -> None:
        # arrange
        crawler = Reddit(subreddit='foo/bar bazz')
        # act
        uri = crawler._get_uri(None)
        # assert
        self.assertEqual(uri, 'https://www.reddit.com/r/foo%2Fbar+bazz.json?after=')

    def test__build_uri_at_after(self) -> None:
        # arrange
        crawler = Reddit(subreddit='test')
        # act
        uri = crawler._get_uri('foobar')
        # assert
        self.assertEqual(uri, 'https://www.reddit.com/r/test.json?after=foobar')

    def test__build_uri_at_after__escape(self) -> None:
        # arrange
        crawler = Reddit(subreddit='test')
        # act
        uri = crawler._get_uri('foo/bar bazz')
        # assert
        self.assertEqual(uri, 'https://www.reddit.com/r/test.json?after=foo%2Fbar+bazz')


class RedditResetTest(unittest.TestCase):

    def test_reset_done(self) -> None:
        # arrange
        crawler = Reddit(subreddit='test')
        crawler._after = 'foo'
        # act
        crawler._reset()
        # assert
        self.assertIsNone(crawler._after)


_FILE_FETCHER = FileFetcher({  # relative to "./testdata_instagram"
    '/r/aww.json?after=': 'aww.json',
}, base_dir=path_join(dirname(__file__), 'testdata_reddit'))


class RedditCrawlTest(unittest.TestCase):

    def setUp(self) -> None:
        self.crawler = Reddit(subreddit='aww')
        self.crawler._remote_fetcher = _FILE_FETCHER

    def tearDown(self) -> None:
        del self.crawler

    def test_crawl(self) -> None:
        # arrange
        expected_after = 't3_dqx42l'
        expected_images = ImageCollection()
        expected_images.add(Image(
            uri='https://i.redd.it/kl3dp9sy5fw31.jpg',
            source='https://www.reddit.com/r/aww/comments/dqx0z4/a_very_photogenic_noodle/'))
        expected_images.add(Image(
            uri='https://i.redd.it/4ltnvj5irdw31.jpg',
            source='https://www.reddit.com/r/aww/comments/dqud6w/3/'))
        expected_images.add(Image(
            uri='https://i.redd.it/nkfjoej8yew31.png',
            source='https://www.reddit.com/r/aww/comments/dqwp8l/left_the_house_for_10_minutes_and_came_back_to/'))
        expected_images.add(Image(
            uri='https://i.redd.it/gcxqswv8igw31.png',
            source='https://www.reddit.com/r/aww/comments/dqz6iz/blind_cutie/'))
        expected_images.add(Image(
            uri='https://i.redd.it/hywobahj9ew31.png',
            source='https://www.reddit.com/r/aww/comments/dqvgm9/i_asked_this_guy_if_he_knocked_over_the_treats/'))
        expected_images.add(Image(
            uri='https://i.redd.it/j4qda3c9scw31.jpg',
            source='https://www.reddit.com/r/aww/comments/dqrxiq/admiral_anchovies_is_two_weeks_old_and_has/'))
        expected_images.add(Image(
            uri='https://i.imgur.com/O2bVClA.jpg',
            source='https://www.reddit.com/r/aww/comments/dqsk7y/meet_our_new_3_month_old_baby_bucko_the_beagle/',
        ))
        expected_images.add(Image(
            uri='https://imgur.com/82LxoWj.jpg',
            source='https://www.reddit.com/r/aww/comments/dqtdo7/im_one_of_a_kind/',
        ))
        # act
        images = self.crawler._crawl()
        # assert
        self.assertEqual(self.crawler._after, expected_after)
        self.assertSetEqual(images, expected_images)
        for expected_image in expected_images:
            for image in images:
                if image == expected_image:
                    # sources are irrelevant for equality, need to be checked manually
                    self.assertEqual(image.source, expected_image.source)


class RedditDescriptionTest(unittest.TestCase):

    def test_description_config(self) -> None:
        # act
        description = Reddit.info()
        # assert
        assert isinstance(description.config, dict)
        for config_key in _reddit_right_config.keys():
            self.assertIn(config_key, description.config)


class RedditLoaderTest(ImageCrawlerLoaderTest):

    @property
    def ic_name(self) -> str:
        return 'Reddit'

    @property
    def ic_class(self) -> Type[BaseImageCrawler]:
        return Reddit

    def test_loader(self) -> None:
        self.check()



import unittest
from typing import Type

from nichtparasoup.imagecrawler import BaseImageCrawler
from nichtparasoup.imagecrawler.picsum import Picsum
from nichtparasoup.testing.imagecrawler import ImageCrawlerLoaderTest

_picsum_right_config = {'width': 800, 'height': 600}


class PicsumConfigCorrect(unittest.TestCase):

    def test__check_config_right_value(self) -> None:
        # arrange
        config_in = _picsum_right_config.copy()
        # act
        config_out = Picsum.check_config(config_in)
        # assert
        self.assertDictEqual(config_in, config_out)


class PicsumConfigWidthTest(unittest.TestCase):

    def setUp(self) -> None:
        self._picsum_right_config_wo_width = _picsum_right_config.copy()
        del self._picsum_right_config_wo_width["width"]

    def tearDown(self) -> None:
        del self._picsum_right_config_wo_width

    def test__check_config_missing_value(self) -> None:
        # assert
        with self.assertRaises(KeyError):
            Picsum.check_config(self._picsum_right_config_wo_width)

    def test__check_config_wrong_type(self) -> None:
        wrong_types = [None, True, "", [], (), {}, self]  # type: ignore
        for wrong_type in wrong_types:
            # arrange
            config_in = self._picsum_right_config_wo_width
            config_in["width"] = wrong_type  # type: ignore
            # assert
            with self.assertRaises(TypeError, msg=repr(config_in)):
                Picsum.check_config(config_in)

    def test__check_config_wrong_value(self) -> None:
        wrong_values = [0, -1]
        for wrong_value in wrong_values:
            # arrange
            config_in = self._picsum_right_config_wo_width
            config_in["width"] = wrong_value
            # assert
            with self.assertRaises(ValueError, msg=repr(config_in)):
                Picsum.check_config(config_in)


class PicsumConfigHeightTest(unittest.TestCase):

    def setUp(self) -> None:
        self._picsum_right_config_wo_height = _picsum_right_config.copy()
        del self._picsum_right_config_wo_height["height"]

    def tearDown(self) -> None:
        del self._picsum_right_config_wo_height

    def test__check_config_missing_value(self) -> None:
        # assert
        with self.assertRaises(KeyError):
            Picsum.check_config(self._picsum_right_config_wo_height)

    def test__check_config_wrong_type(self) -> None:
        wrong_types = [None, True, "", [], (), {}, self]  # type: ignore
        for wrong_type in wrong_types:
            # arrange
            config_in = self._picsum_right_config_wo_height
            config_in["height"] = wrong_type  # type: ignore
            # assert
            with self.assertRaises(TypeError, msg=repr(config_in)):
                Picsum.check_config(config_in)

    def test__check_config_wrong_value(self) -> None:
        wrong_values = [0, -1]
        for wrong_value in wrong_values:
            # arrange
            config_in = self._picsum_right_config_wo_height
            config_in["height"] = wrong_value
            # assert
            with self.assertRaises(ValueError, msg=repr(config_in)):
                Picsum.check_config(config_in)


class PicsumCrawlTest(unittest.TestCase):

    def test_crawl(self) -> None:
        # arrange
        crawler = Picsum(**_picsum_right_config)
        # act
        images_crawled = crawler.crawl()
        images_crawled_len = len(images_crawled)
        image_crawled = images_crawled.pop() if images_crawled_len else None
        # assert
        self.assertEqual(images_crawled_len, crawler._bunch, "crawler did not finish")
        if image_crawled:
            self.assertTrue(image_crawled.is_generic, 'this is not generic')


class PicsumDescriptionTest(unittest.TestCase):

    def test_description_config(self) -> None:
        # act
        description = Picsum.info()
        # assert
        assert isinstance(description.config, dict)
        for config_key in _picsum_right_config.keys():
            self.assertIn(config_key, description.config)


class PicsumLoaderTest(ImageCrawlerLoaderTest):

    @property
    def ic_name(self) -> str:
        return 'Picsum'

    @property
    def ic_class(self) -> Type[BaseImageCrawler]:
        return Picsum

    def test_loader(self) -> None:
        self.check()



__all__ = ["DummyImage"]

from typing import Any, Dict

from nichtparasoup.imagecrawler import BaseImageCrawler, Image, ImageCollection, ImageCrawlerConfig, ImageCrawlerInfo


class DummyImage(BaseImageCrawler):

    @classmethod
    def info(cls) -> ImageCrawlerInfo:
        from textwrap import dedent
        return ImageCrawlerInfo(
            description='Dynamic Dummy Image from https://dummyimage.com',
            long_description=dedent('''
                Not an actual crawler.
                Just a thing that delivers images that are useful to test the web UI of nichtparasoup.
                ''').strip(),
            config=dict(
                width='width of a test image',
                height='height of a test image',
            ),
            icon_url='https://dummyimage.com/favicon.ico',
        )

    @classmethod
    def check_config(cls, config: Dict[Any, Any]) -> ImageCrawlerConfig:
        width = config["width"]
        height = config["height"]
        if type(width) is not int:
            raise TypeError("width {!r} is not int".format(width))
        if type(height) is not int:
            raise TypeError("height {!r} is not int".format(height))
        if width <= 0:
            raise ValueError("width {!r} <= 0".format(width))
        if height <= 0:
            raise ValueError("height {!r} <= 0".format(height))
        return ImageCrawlerConfig(
            width=width,
            height=height,
        )

    # trailing '?' is needed to prevent random technical issues with the generator
    _IMAGE_URI_TEMPLATE = 'https://dummyimage.com/{}x{}/567/fed/?'

    @classmethod
    def _get_image_uri(cls, width: int, height: int) -> str:
        return cls._IMAGE_URI_TEMPLATE.format(width, height)

    def _reset(self) -> None:
        # nothing to reset since this is a generic crawler
        pass

    _BUNCH = 10

    def _crawl(self) -> ImageCollection:
        images = ImageCollection()
        for _ in range(self._BUNCH):
            images.add(Image(
                uri=self._get_image_uri(**self._config),
                source='https://dummyimage.com/',
                is_generic=True,
            ))
        return images



import unittest
from os.path import dirname, join as path_join
from typing import Any, Dict, Optional, Type
from urllib.parse import parse_qs, urlsplit

from ddt import data as ddt_data, ddt, idata as ddt_idata, unpack as ddt_unpack  # type: ignore

from nichtparasoup.imagecrawler import BaseImageCrawler, Image, ImageCollection
from nichtparasoup.imagecrawler.pr0gramm import Pr0gramm
from nichtparasoup.testing.imagecrawler import FileFetcher, ImageCrawlerLoaderTest


@ddt
class Pr0grammConfigPromotedTest(unittest.TestCase):

    @ddt_data((True,), (False,))  # type: ignore
    @ddt_unpack  # type: ignore
    def test__check_value(self, promoted: bool) -> None:
        # arrange
        config_in = dict(promoted=promoted)
        # act
        config_out = Pr0gramm.check_config(config_in)
        # assert
        self.assertEqual(promoted, config_out['promoted'])

    def test__check_missing_value_defaults_to_true(self) -> None:
        # arrange
        config_in = dict()  # type: Dict[Any, Any]
        # act
        config_out = Pr0gramm.check_config(config_in)
        # assert
        self.assertTrue(config_out['promoted'])

    def test__check_wrong_type(self) -> None:
        wrong_types = ['', 23, 4.2, [], (), {}, self]
        for wrong_type in wrong_types:
            # assert
            with self.assertRaises(TypeError, msg=repr(wrong_type)):
                Pr0gramm.check_config(dict(promoted=wrong_type))


@ddt
class Pr0grammConfigTagsTest(unittest.TestCase):

    def test__check_missing_value(self) -> None:
        # arrange
        config_in = dict()  # type: Dict[Any, Any]
        # act
        config_out = Pr0gramm.check_config(config_in)
        # assert
        self.assertIsNone(config_out['tags'])

    @ddt_data(('!"test"',))  # type: ignore
    @ddt_unpack  # type: ignore
    def test__check_value(self, tags: str) -> None:
        # arrange
        config_in = dict(tags=tags)
        # act
        config_out = Pr0gramm.check_config(config_in)
        # assert
        self.assertEqual(tags, config_out['tags'])

    def test__check_wrong_tag_value(self) -> None:
        wrong_values = ['', ' \n\t', ' !', '! \n\t']
        for wrong_value in wrong_values:
            # assert
            with self.assertRaises(ValueError, msg=repr(wrong_value)):
                Pr0gramm.check_config(dict(tags=wrong_value))

    def test__check_wrong_type(self) -> None:
        wrong_types = [True, 23, 4.2, [], (), {}, self]
        for wrong_type in wrong_types:
            # assert
            with self.assertRaises(TypeError, msg=repr(wrong_type)):
                Pr0gramm.check_config(dict(tags=wrong_type))


@ddt
class Pr0grammUrlBuilderTest(unittest.TestCase):

    def test_base(self) -> None:
        # act
        api_uri = Pr0gramm._get_api_uri(flags=0, promoted=False)
        (scheme, netloc, path, _, _) = urlsplit(api_uri)
        # assert
        self.assertEqual('https', scheme)
        self.assertEqual('pr0gramm.com', netloc)
        self.assertEqual('/api/items/get', path)

    @ddt_idata((i, str(i)) for i in range(1, 16, 1))  # type: ignore
    @ddt_unpack  # type: ignore
    def test_flags(self, flags: int, flags_qs: str) -> None:
        # act
        api_uri = Pr0gramm._get_api_uri(flags=flags, promoted=False)
        (_, _, _, query_string, _) = urlsplit(api_uri)
        query = parse_qs(query_string)
        # assert
        self.assertEqual([flags_qs], query['flags'])

    @ddt_data((True, '1'), (False, '0'))  # type: ignore
    @ddt_unpack  # type: ignore
    def test_promoted(self, promoted: bool, promoted_qs: str) -> None:
        # act
        api_uri = Pr0gramm._get_api_uri(flags=0, promoted=promoted)
        (_, _, _, query_string, _) = urlsplit(api_uri)
        query = parse_qs(query_string)
        # assert
        self.assertEqual([promoted_qs], query['promoted'])

    @ddt_data((None, '! -"video"'), ('', '! -"video"'), ('!"test"', '!("test") -"video"'))  # type: ignore
    @ddt_unpack  # type: ignore
    def test_tags(self, tags: Optional[str], tags_qs: str) -> None:
        # act
        api_uri = Pr0gramm._get_api_uri(flags=0, promoted=False, tags=tags)
        (_, _, _, query_string, _) = urlsplit(api_uri)
        query = parse_qs(query_string)
        # assert
        self.assertEqual([tags_qs], query['tags'])

    @ddt_data((None, None), (23, '23'))  # type: ignore
    @ddt_unpack  # type: ignore
    def test_older(self, older: Optional[int], older_qs: str) -> None:
        # act
        api_uri = Pr0gramm._get_api_uri(flags=0, promoted=False, older=older)
        (_, _, _, query_string, _) = urlsplit(api_uri)
        query = parse_qs(query_string)
        # assert
        self.assertEqual([older_qs], query.get('older', [None]))


_FILE_FETCHER = FileFetcher({  # relative to "./testdata_pr0gramm"
    '/api/items/get?flags=1&promoted=1&tags=%21%28s%3A15000%29+-%22video%22':
        'get-flags_1-promoted_1-tags_s15000-video.json',
    '/api/items/get?flags=1&promoted=1&tags=%21%28s%3A1000%29+-%22video%22':
        'get-flags_1-promoted_1-tags_s1000-video.json',
    '/api/items/get?flags=1&promoted=0&tags=%21%28s%3A1000%29+-%22video%22':
        'get-flags_1-promoted_0-tags_s1000-video.json',
}, base_dir=path_join(dirname(__file__), 'testdata_pr0gramm'))


class Pr0grammResetTest(unittest.TestCase):

    def test_reset_done(self) -> None:
        # arrange
        crawler = Pr0gramm()
        crawler._older = 1337
        # act
        crawler._reset()
        # assert
        self.assertIsNone(crawler._older)


@ddt
class Pr0grammCrawlTest(unittest.TestCase):

    def test_reset_at_end(self) -> None:
        # arrange
        crawler = Pr0gramm(promoted=True, tags='!s:15000')
        crawler._remote_fetcher = _FILE_FETCHER
        crawler._reset_before_next_crawl = False
        # act
        crawler.crawl()
        # assert
        self.assertTrue(crawler._reset_before_next_crawl)

    def test_no_reset_before_end(self) -> None:
        # arrange
        crawler = Pr0gramm(flags=1, promoted=True, tags='!s:1000')
        crawler._remote_fetcher = _FILE_FETCHER
        crawler._reset_before_next_crawl = False
        # act
        crawler.crawl()
        # assert
        self.assertFalse(crawler._reset_before_next_crawl)

    @ddt_data((True, 503528), (False, 3652675))  # type: ignore
    @ddt_unpack  # type: ignore
    def test_crawl_cursor(self, promoted: bool, expected_cursor: Optional[int]) -> None:
        # arrange
        crawler = Pr0gramm(flags=1, promoted=promoted, tags='!s:1000')
        crawler._remote_fetcher = _FILE_FETCHER
        crawler._older = None
        # act
        crawler._crawl()
        # assert
        self.assertEqual(crawler._older, expected_cursor, msg='promoted={!r}'.format(promoted))

    def test_crawl_images(self) -> None:
        # arrange
        crawler = Pr0gramm(flags=1, promoted=True, tags='!s:15000')
        crawler._remote_fetcher = _FILE_FETCHER
        expected_images = ImageCollection()
        expected_images.add(Image(
            uri='https://img.pr0gramm.com/2018/10/11/ac41a1fbcc3abf09.png',
            source='https://pr0gramm.com/new/2782197',
        ))
        expected_images.add(Image(
            uri='https://img.pr0gramm.com/2015/07/05/5624d30ec6e743b2.png',
            source='https://pr0gramm.com/new/879293',
        ))
        # act
        images = crawler._crawl()
        # assert
        self.assertSetEqual(images, expected_images)
        for expected_image in expected_images:
            for image in images:
                if image == expected_image:
                    # sources are irrelevant for equality, need to be checked manually
                    self.assertEqual(image.source, expected_image.source)


class Pr0grammDescriptionTest(unittest.TestCase):

    def test_description_config(self) -> None:
        # act
        description = Pr0gramm.info()
        # assert
        assert isinstance(description.config, dict)
        self.assertIn('promoted', description.config)
        self.assertIn('tags', description.config)


class Pr0grammLoaderTest(ImageCrawlerLoaderTest):

    @property
    def ic_name(self) -> str:
        return 'Pr0gramm'

    @property
    def ic_class(self) -> Type[BaseImageCrawler]:
        return Pr0gramm

    def test_loader(self) -> None:
        self.check()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# BASED ON https://github.com/navdeep-G/setup.py

import io
import os
import sys

from setuptools import find_packages, setup

CURRENT_PYTHON = sys.version_info[:2]
REQUIRED_PYTHON = (3, 5)

# This check and everything above must remain compatible with Python 2.7.
if CURRENT_PYTHON < REQUIRED_PYTHON:
    import textwrap

    sys.stderr.write(textwrap.dedent('''
    UNSUPPORTED PYTHON VERSION

    This version of nichtparasoup requires Python {py_required[0]}.{py_required[1]},
    but you're trying to install it on Python {py_current[0]}.{py_current[1]}.

    '''.format(py_current=CURRENT_PYTHON, py_required=REQUIRED_PYTHON)))
    sys.exit(1)

# Package meta-data.
NAME = 'nichtparasoup'
DESCRIPTION = 'A hackspaces entertainment system'
URL = 'https://github.com/k4cg/nichtparasoup/'
AUTHOR = 'K4CG'
EMAIL = 'info@k4cg.org'
REQUIRES_PYTHON = '>={}.{}'.format(*REQUIRED_PYTHON)

URLS = dict(
    Source='https://github.com/k4cg/nichtparasoup/',
    Documentation='https://github.com/k4cg/nichtparasoup/tree/master/docs',
)

# What packages are required for this module to be executed?
REQUIRED = [
    'typing-extensions>=3.7.4;python_version<"3.8"',  # for `_internals`, `imagecrawler/instagram`
    'ruamel.yaml>=0.16', 'yamale>=2.0',  # for `config`
    'werkzeug>=0.15',  # for `webserver`
    'mako>=1.1',  # for `webserver`'s templates
    'setuptools>=40.0',  # for imagecrawler-plugin-architecture
]

# What packages are optional?
EXTRAS = dict(
    colors=[
        'termcolor>=1.1',
    ],
    development=[
        'tox>=3.14',
        'isort>=4.3',
    ],
    testing=[
        'flake8',
        'flake8-annotations;python_version>="3.6"',
        'flake8-bugbear',
        'flake8-isort',
        'flake8-pep3101',
        'pep8-naming',
        'mypy',
        'coverage',
        'pytest',
        'ddt',
        # 'flake8-builtins',  # nice in general, but seams not bug-free, yet.
        # 'lake8-docstrings', # not in use, until all docs are done.
        # 'flake8-quotes',    # not in use until black code styles are setup
    ],
)

# The rest you shouldn't have to touch too much :)
# ------------------------------------------------
# Except, perhaps the License and Trove Classifiers!
# If you do change the License, remember to change the Trove Classifier for that!

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!
try:
    with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
        long_description = '\n' + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION

# Where the magic happens:
setup(
    name=NAME,
    use_scm_version=dict(
        root=here,
        fallback_version='UNKNOWN.scm',
        write_to=os.path.join('nichtparasoup', '__version__.py'),
    ),
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    project_urls=URLS,
    packages=find_packages(exclude=['tests', '*.tests', '*.tests.*', 'tests.*', 'examples']),
    entry_points=dict(
        console_scripts=[
            'nichtparasoup = nichtparasoup.cmdline:main',
        ],
    ),
    setup_requires=['setuptools', 'setuptools_scm>=3.3.3'],
    install_requires=REQUIRED,
    extras_require=EXTRAS,
    include_package_data=True,
    license='MIT',
    classifiers=[
        # Trove classifiers - https://packaging.python.org/specifications/core-metadata/#metadata-classifier
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Games/Entertainment',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Typing :: Typed',
    ],
    keywords=[
        'image-crawler',
        'instagram',
        'pr0gramm',
        'reddit',
    ],
)
